# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
Functions from MNE-Python that have been removed from public API
"""
import numpy as np


##############################################################################
# formerly in mne.source_space

_src_kind_dict = {
    "vol": "volume",
    "surf": "surface",
    "discrete": "discrete",
}


def label_src_vertno_sel(label, src):
    """Find vertex numbers and indices from label.

    Parameters
    ----------
    label : Label
        Source space label.
    src : dict
        Source space.

    Returns
    -------
    vertices : list of length 2
        Vertex numbers for lh and rh.
    src_sel : array of int (len(idx) = len(vertices[0]) + len(vertices[1]))
        Indices of the selected vertices in sourse space.
    """
    if src[0]["type"] != "surf":
        raise ValueError(
            "Labels are only supported with surface source spaces, "
            f"got {_src_kind_dict[src[0]['type']]} source space"
        )

    vertno = [src[0]["vertno"], src[1]["vertno"]]

    if label.hemi == "lh":
        vertno_sel = np.intersect1d(vertno[0], label.vertices)
        src_sel = np.searchsorted(vertno[0], vertno_sel)
        vertno[0] = vertno_sel
        vertno[1] = np.array([], int)
    elif label.hemi == "rh":
        vertno_sel = np.intersect1d(vertno[1], label.vertices)
        src_sel = np.searchsorted(vertno[1], vertno_sel) + len(vertno[0])
        vertno[0] = np.array([], int)
        vertno[1] = vertno_sel
    elif label.hemi == "both":
        vertno_sel_lh = np.intersect1d(vertno[0], label.lh.vertices)
        src_sel_lh = np.searchsorted(vertno[0], vertno_sel_lh)
        vertno_sel_rh = np.intersect1d(vertno[1], label.rh.vertices)
        src_sel_rh = np.searchsorted(vertno[1], vertno_sel_rh) + len(vertno[0])
        src_sel = np.hstack((src_sel_lh, src_sel_rh))
        vertno = [vertno_sel_lh, vertno_sel_rh]
    else:
        raise Exception("Unknown hemisphere type")

    return vertno, src_sel



import numpy as np
from scipy import ndimage
import mne
from eelbrain._stats import testnd as tnd
from eelbrain import *

mne.set_log_level('warning')


ds = datasets.get_mne_sample(-0.1, 1.0, src='ico', sub="[0]")
src = ds['src']
bin_map = (src.abs()[0] > 1).x
out = np.empty(bin_map.shape, np.uint32)
struct = ndimage.generate_binary_structure(2, 1)
struct[::2] = False
conn = src.source.adjacency()
criteria = None

print("tnd._label_clusters_binary(bin_map, out, struct, False, bin_map.shape, conn, criteria)")



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: sphinx
#       format_version: '1.1'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
.. _exa-boxplot:

Boxplot
=======

Much of the functionality is shared with :class:`plot.Barplot`.
"""
from eelbrain import *

###############################################################################
# Basic boxplot
# ^^^^^^^^^^^^^

ds = datasets.get_uv()
ds.summary()
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds)

###############################################################################
# Test against population mean
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Instead of pairwise tests, test the values in each category against a population mean. Use the ``test`` parameter to set the value to test against. In the example, ``test=0`` will test whether each sample is significantly different form zero. Use the ``tail`` parameter to determine the tailedness of the test (``tail=1`` to test for a value greater than 0).

p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, test=0, tail=1)

###############################################################################
# Appearance: colored boxes and category labels
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

colors = {
    ('a1', 'b1'): (1.0, 0.0, 0.0),
    ('a1', 'b2'): (1.0, 0.7, 0.0),
    ('a2', 'b1'): (0.0, 0.0, 1.0),
    ('a2', 'b2'): (0.0, 0.8, 1.0),
}
labels = {
    ('a1', 'b1'): 'A-1 (B-1)',
    ('a1', 'b2'): 'A-1 (B-2)',
    ('a2', 'b1'): 'A-2 (B-1)',
    ('a2', 'b2'): 'A-2 (B-2)',
}
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, colors=colors, xticks=labels)

colors = {
    ('a1', 'b1'): plot.Style((1, .5, .5), hatch=''),
    ('a1', 'b2'): plot.Style((1, .5, .5), hatch='///'),
    ('a2', 'b1'): plot.Style((.5, .5, 1), hatch=''),
    ('a2', 'b2'): plot.Style((.5, .5, 1), hatch='///'),
}
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, colors=colors)

colors = {
    ('a1', 'b1'): plot.Style('w', hatch='///'),
    ('a1', 'b2'): plot.Style('w', hatch='/'),
    ('a2', 'b1'): plot.Style('w', hatch='O'),
    ('a2', 'b2'): plot.Style('w', hatch='.'),
}
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, colors=colors)

###############################################################################
# Label fliers
# ^^^^^^^^^^^^
#
# The ``label_fliers=True`` option is used to identify outlier points (labels are based on the ``match`` argument).

p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, label_fliers=True)



"""
.. _exa-colormaps:

Eelbrain Colormaps
==================

Eelbrain adds some colormaps to `matplotlib's standard colormaps
<https://matplotlib.org/gallery/color/colormap_reference.html>`_.
In general, the colormaps come with an ``*-a`` ("alpha") version which uses
transparency for low values instead of a solid color. For example, to use the
``polar`` map with transparency, use ``polar-a``.
"""
import numpy as np
import matplotlib.pyplot as plt


gradient = np.linspace(0, 1, 256)
gradient = np.vstack((gradient, gradient))

cmap_list = [
    'polar',
    'xpolar',
    'lpolar',
    'lux',
    'lux-gray',
    'polar-lux',
    'lux-purple',
    'polar-lux-purple',
]

nrows = len(cmap_list)
figh = 0.35 + 0.15 + (nrows + (nrows-1)*0.1)*0.22
fig, axs = plt.subplots(nrows=nrows, figsize=(6.4, figh))
fig.subplots_adjust(top=1-.35/figh, bottom=.15/figh, left=0.2, right=0.99)

axs[0].set_title('Eelbrain colormaps', fontsize=14)

for ax, name in zip(axs, cmap_list):
    ax.imshow(gradient, aspect='auto', cmap=name)
    ax.text(-.01, .5, name, va='center', ha='right', fontsize=10,
            transform=ax.transAxes)

for ax in axs:
    ax.set_axis_off()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Wrapper for binaries

For permission errors: try ``os.chmod`` or ``os.chown``


subprocess documentation
------------------------

http://docs.python.org/library/subprocess.html
http://www.doughellmann.com/PyMOTW/subprocess/
http://explanatorygap.net/2010/05/10/python-subprocess-over-shell/
http://codeghar.wordpress.com/2011/12/09/introduction-to-python-subprocess-module/
"""
import fnmatch
import logging
import os
import subprocess

import mne
from mne.utils import get_subjects_dir, run_subprocess

from . import ui


# environment variables that can hold the path to this tool
_env_vars = {'mne': 'MNE_ROOT',
             'freesurfer': 'FREESURFER_HOME',
             'edfapi': 'EYELINK_HOME'}

# bin directory relative to tool location
_bin_dirs = {'mne': 'bin',
             'freesurfer': 'bin',
             'edfapi': ''}

# binaries to look for to test whether the tool directory is set correctly
_test_bins = {'mne': ('mne_add_patch_info', 'mne_analyze'),
              'freesurfer': ('mri_annotation2label',),
              'edfapi': ('edf2asc',)}


def get_root(package):
    "Get the root path for a package"
    if package not in _env_vars:
        raise KeyError("Unknown binary package: %r" % package)
    root = mne.get_config(_env_vars[package])
    if root:
        test_root(package, root)
    else:
        root = _ask_user_for_bin_dir(package)
    return root


def get_bin(package, name):
    "Get path for a binary"
    root = get_root(package)
    bin_path = os.path.join(root, _bin_dirs[package], name)
    if os.path.exists(bin_path):
        return bin_path
    else:
        raise IOError("Binary does not exist: " + bin_path)


def test_root(package, root):
    bin_dir = os.path.join(root, _bin_dirs[package])
    return all(os.path.exists(os.path.join(bin_dir, f)) for f in _test_bins[package])


def _ask_user_for_bin_dir(package):
    """
    Change the location from which binaries are used.

    Parameters
    ----------
    package : str
        Binary package for which to set the directory. One from:
        ``['mne', 'freesurfer', 'edfapi']``
    """
    title = "Select %s Directory" % package
    message = "Please select the directory of the %s package." % package
    while True:
        answer = ui.ask_dir(title, message, must_exist=True)
        if answer:
            if test_root(package, answer):
                mne.set_config(_env_vars[package], answer)
                return answer
            else:
                ui.message("Wrong Directory", "This is not the right directory.")
        else:
            raise IOError("%s directory not set" % package)


_verbose = 1


def command_exists(cmd):
    """Return True if the corresponding command exists

    based http://stackoverflow.com/a/11069822/166700
    """
    return subprocess.call(["type", cmd], stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE) == 0


def open_in_finder(path):
    """Open ``path`` in the finder"""
    os.system('open %s' % path)


def get_fs_home():
    fs_home = mne.get_config('FREESURFER_HOME')
    save_fs_home = False
    while True:
        problem = fs_home_problem(fs_home)
        if problem:
            save_fs_home = True
            print(problem)
            if fs_home == os.environ.get('FREESURFER_HOME', 0):
                print("WARNING: This directory is set as FREESURFER_HOME "
                      "environment variable. As long as you don't remove this "
                      "environment variable, you will be asked for the proper "
                      "FreeSurfer location every time a FreeSurfer command is "
                      "run.")
            message = "Please select the directory where FreeSurfer is installed"
            print(message)
            fs_home = ui.ask_dir("Select FreeSurfer Directory", message)
            if fs_home is False:
                raise RuntimeError("Could not find FreeSurfer")
        else:
            break
    if save_fs_home:
        mne.set_config('FREESURFER_HOME', fs_home)
    return fs_home


def run_freesurfer_command(command, subjects_dir):
    "Run a FreeSurfer command"
    env = os.environ.copy()
    env['SUBJECTS_DIR'] = subjects_dir
    env['FREESURFER_HOME'] = fs_home = get_fs_home()
    bin_path = os.path.join(fs_home, 'bin')
    if bin_path not in env['PATH']:
        env['PATH'] = ':'.join((bin_path, env['PATH']))

    # run command
    run_subprocess(command, env=env)


def fs_home_problem(fs_home):
    """Check FREESURFER_HOME path

    Return str describing problem or None if the path is okay.
    """
    if fs_home is None:
        return "The FreeSurfer path is not set"
    elif not os.path.exists(fs_home):
        return "The FreeSurfer path does not exist: %s" % fs_home
    else:
        test_path = os.path.join(fs_home, 'bin', 'mri_surf2surf')
        if not os.path.exists(test_path):
            return ("The FreeSurfer path is invalid (%s does not contain bin/"
                    "mri_surf2surf)" % fs_home)


def _run(cmd, v=None, cwd=None, block=True):
    """Run a command

    Parameters
    ----------
    cmd: list of strings
        command that is submitted to subprocess.Popen.
    v : 0 | 1 | 2 | None
        verbosity level (0: nothing;  1: stderr;  2: stdout;  None: use
        _verbose module attribute)

    """
    if v is None:
        v = _verbose

    if v > 1:
        print("> COMMAND:")
        for line in cmd:
            print(repr(line))

    sp = subprocess.Popen(list(map(str, cmd)), cwd=cwd,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if block:
        stdout, stderr = sp.communicate()

        if v > 1:
            print("\n> stdout:")
            print(stdout)

        if v > 0 and stderr:
            print('\n> stderr:')
            print(stderr)

        return stdout, stderr


def setup_mri(subject, subjects_dir=None, ico=4, block=False, redo=False):
    """Prepare an MRI for use in the mne-pipeline

    Performes the following steps:

     - runs mne_setup_mri
     - runs mne_setup_source_space
     - runs mne_watershed_bem
     - creates symlinks for watershed files in the bem directory
     - runs mne_setup_forward_model (see MNE manual section 3.7, p. 25)


    Parameters
    ----------
    subject : str
        Subject whose MRI should be processed.
    subjects_dir : None | str(path)
        Overrides `sys.environ['SUBJECTS_DIR']` if not `None`.
    ico : int
        `--ico` argument for `mne_setup_forward_model`.
    block : bool
        Block the Python interpreter until mne_setup_forward_model is
        finished.
    redo : bool
        Run the commands even if the target files already exist.


    .. note::
        The utility needs permission to access the MRI files. In case of a
        permission error, the following shell command can be used on the mri
        folder to set permissions appropriately::

            $ sudo chmod -R 7700 mri-folder

    """
    if subjects_dir is None:
        subjects_dir = os.environ['SUBJECTS_DIR']
        _sub_dir = None
    else:
        _sub_dir = os.environ.get('SUBJECTS_DIR', None)
        os.environ['SUBJECTS_DIR'] = subjects_dir

    bemdir = os.path.join(subjects_dir, subject, 'bem')

    # mne_setup_mri
    tgt = os.path.join(subjects_dir, subject, 'mri', 'T1', 'neuromag', 'sets',
                       'COR.fif')
    if redo or not os.path.exists(tgt):
        cmd = [get_bin('mne', 'mne_setup_mri'),
               '--subject', subject,
               '--mri', 'T1']
        _run(cmd)

    # mne_setup_source_space
    tgt = os.path.join(bemdir, '%s-ico-%i-src.fif' % (subject, ico))
    if redo or not os.path.exists(tgt):
        cmd = [get_bin('mne', 'mne_setup_source_space'),
               '--subject', subject,
               '--ico', ico]
        _run(cmd)

    # mne_watershed_bem
    cmd = [get_bin('mne', 'mne_watershed_bem'), '--subject', subject]
    if redo:
        cmd.append('--overwrite')
    _run(cmd)

    # symlinks (MNE-manual 3.6, p. 24)
    for name in ['inner_skull', 'outer_skull', 'outer_skin']:
        src = os.path.join('watershed', '%s_%s_surface' % (subject, name))
        dest = os.path.join(bemdir, '%s.surf' % name)
        if os.path.exists(dest):
            if os.path.islink(dest):
                abs_src = os.path.join(bemdir, src)
                if os.path.realpath(dest) == abs_src:
                    pass
                else:
                    logging.debug("replacing symlink: %r" % dest)
                    os.remove(dest)
                    os.symlink(src, dest)
                    # can raise an OSError: [Errno 13] Permission denied
            else:
                raise IOError("%r exists and is no symlink" % dest)
        else:
            logging.debug("creating symlink: %r" % dest)
            os.symlink(src, dest)

    # mne_setup_forward_model
    cmd = [get_bin('mne', "mne_setup_forward_model"),
           '--subject', subject,
           '--surf',
           '--ico', ico,
           '--homog']

    _run(cmd, block=block)  # -> creates a number of files in <mri_sdir>/bem
    if _sub_dir:
        os.environ['SUBJECTS_DIR'] = _sub_dir


def _run_mne_gui(name, cwd, modal, subject, subjects_dir):
    root = get_root('mne')
    env = os.environ.copy()
    env['MNE_ROOT'] = root
    if subjects_dir is not None:
        env['SUBJECTS_DIR'] = subjects_dir
    if subject:
        env['SUBJECT'] = subject

    setup_path = os.path.join(root, 'bin', 'mne_setup_sh').replace(' ', r'\ ')
    setup = '. %s' % setup_path

    cmd = os.path.join(root, 'bin', name).replace(' ', r'\ ')
    p = subprocess.Popen(setup + ';' + cmd, shell=True, cwd=cwd, env=env)

    if modal:
        print("Waiting for %s to be closed..." % name)
        p.wait()


def run_mne_analyze(fif_dir, subject=None, subjects_dir=None, modal=False):
    """Invoke mne_analyze (e.g., for manual coregistration)

    Parameters
    ----------
    fif_dir : str
        the directory containing the fiff files.
    subject : None | str
        The name of the MRI subject.
    subjects_dir : None | str
        Override the SUBJECTS_DIR environment variable.
    modal : bool
        Causes the shell to block until mne_analyze is closed.


    Notes
    -----
    **Coregistration Procedure** (For more information see MNE-manual 3.10 &
    12.11):

    #. File > Load Surface: select the subject`s directory and "Inflated"
    #. File > Load digitizer data: select the fiff file
    #. View > Show viewer

       a. ``Options``: Switch cortical surface display off, make
          scalp transparent. ``Done``

    #. Adjust > Coordinate Alignment:

       a. set LAP, Nasion and RAP.
       b. ``Align using fiducials``.
       c. (``Omit``)
       d. Run ``ICP alignment`` with 20 steps
       e. ``Save default``

    this creates a file next to the raw file with the '-trans.fif' extension.

    """
    _run_mne_gui('mne_analyze', fif_dir, modal, subject, subjects_dir)


def run_mne_browse_raw(fif_dir, subject=None, subjects_dir=None, modal=False):
    """Invoke mne_browse_raw

    Parameters
    ----------
    fif_dir : str
        the directory containing the fiff files.
    subject : None | str
        The name of the MRI subject.
    subjects_dir : None | str
        Override the SUBJECTS_DIR environment variable.
    modal : bool
        Causes the shell to block until mne_browse_raw is closed.
    """
    _run_mne_gui('mne_browse_raw', fif_dir, modal, subject, subjects_dir)


# freesurfer---


def _fs_hemis(arg):
    if arg == '*':
        return ['lh', 'rh']
    elif arg in ['lh', 'rh']:
        return [arg]
    else:
        raise ValueError("hemi has to be 'lh', 'rh', or '*' (no %r)" % arg)


def _fs_subjects(arg, exclude=[], subjects_dir=None):
    if '*' in arg:
        subjects_dir = get_subjects_dir(subjects_dir)
        subjects = fnmatch.filter(os.listdir(subjects_dir), arg)
        subjects = list(filter(os.path.isdir, subjects))
        for subject in exclude:
            if subject in subjects:
                subjects.remove(subject)
    else:
        subjects = [arg]
    return subjects



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False, language_level=3

from cython.parallel cimport prange
from libc.stdlib cimport malloc, free

cimport numpy as np


ctypedef np.float64_t FLOAT64
ctypedef np.int64_t INT64


def convolve_2d(
        FLOAT64 [:,:,:] h_flat,  # (n_h_only, n_shared, n_h_times)
        FLOAT64 [:,:,:] x_flat,  # (n_x_only, n_shared, n_x_times)
        FLOAT64 [:,:] x_pads,  # (n_x_only, n_shared)
        int h_i_start,  # offset of the first sample of h
        INT64 [:,:] segments,  # (n_segments, 2)
        FLOAT64 [:,:,:] out_flat,  # (n_x_only, n_h_only, n_x_times)
):
    cdef:
        Py_ssize_t i_h, i_x, i_out
        Py_ssize_t n_x_only = x_flat.shape[0]
        Py_ssize_t n_h_only = h_flat.shape[0]

    # loop through x and h dimensions
    for i_out in prange(n_h_only * n_x_only , nogil=True, schedule='guided'):
        i_h = i_out // n_x_only
        i_x = i_out % n_x_only
        convolve_1d(h_flat[i_h], x_flat[i_x], x_pads[i_x], h_i_start, segments, out_flat[i_x, i_h])


cpdef int convolve_1d(
        FLOAT64 [:,:] h,  # (n_x, n_h_times)
        FLOAT64 [:,:] x,  # (n_x, n_x_times)
        FLOAT64 [:] x_pads,  # (n_x,)
        int h_i_start,  # offset of the first sample of h
        INT64 [:,:] segments,  # (n_segments, 2)
        FLOAT64 [:] out,  # (n_x_times,)
) noexcept nogil:
    cdef:
        Py_ssize_t i, i_h, i_t, i_tau, i_x, start, stop
        Py_ssize_t n_x = h.shape[0]
        Py_ssize_t h_n_times = h.shape[1]
        Py_ssize_t h_i_stop = h_i_start + h_n_times
        Py_ssize_t pad_head_n_times = max(0, h_n_times + h_i_start - 1)
        Py_ssize_t pad_tail_n_times = -min(0, h_i_start)
        FLOAT64 * out_pad
        FLOAT64 * pad_head
        FLOAT64 * pad_tail

    # padding
    if pad_head_n_times or pad_tail_n_times:
        # out_pad: response to padding value
        out_pad = <FLOAT64*> malloc(sizeof(FLOAT64) * h_n_times)
        for i_t in range(h_n_times):
            out_pad[i_t] = 0
            for i_x in range(n_x):
                out_pad[i_t] += x_pads[i_x] * h[i_x, i_t]
        # padding for pre-
        if pad_head_n_times:
            pad_head = <FLOAT64*> malloc(sizeof(FLOAT64) * pad_head_n_times)
            for i_t in range(pad_head_n_times):
                pad_head[i_t] = 0
                for i_tau in range(min(pad_head_n_times - i_t, h_n_times)):
                    pad_head[i_t] += out_pad[h_n_times - i_tau - 1]
        # padding for post-
        if pad_tail_n_times:
            pad_tail = <FLOAT64*> malloc(sizeof(FLOAT64) * pad_tail_n_times)
            for i_t in range(pad_tail_n_times):
                pad_tail[i_t] = 0
                for i_tau in range(min(i_t+1, h_n_times)):
                    pad_tail[i_t] += out_pad[i_tau]

    for i in range(segments.shape[0]):
        start = segments[i, 0]
        stop = segments[i, 1]
        out[start: stop] = 0
        if pad_head_n_times:
            for i_tau in range(pad_head_n_times):
                out[start + i_tau] += pad_head[i_tau]
        if pad_tail_n_times:
            for i_tau in range(pad_tail_n_times):
                out[stop - pad_tail_n_times + i_tau] += pad_tail[i_tau]
        convolve_segment(h, x[:, start:stop], out[start:stop], h_i_start, h_i_stop)

    if pad_head_n_times or pad_tail_n_times:
        free(out_pad)
        if pad_head_n_times:
            free(pad_head)
        if pad_tail_n_times:
            free(pad_tail)

    return 0


cdef int convolve_segment(
        FLOAT64 [:,:] h,  # n_x, n_h_times
        FLOAT64 [:,:] x,  # n_x, n_x_times
        FLOAT64 [:] out,  # n_x_times
        Py_ssize_t i_start,
        Py_ssize_t i_stop,
) noexcept nogil:
    cdef:
        Py_ssize_t i_x, i_t, i_tau, i_t_tau
        Py_ssize_t n_times = x.shape[1]
        Py_ssize_t n_x = x.shape[0]
        Py_ssize_t n_tau = i_stop - i_start

    for i_t in range(n_times):
        for i_tau in range(n_tau):
            for i_x in range(n_x):
                i_t_tau = i_t + i_start + i_tau
                if i_t_tau < 0 or i_t_tau >= n_times:
                    continue
                out[i_t_tau] += h[i_x, i_tau] * x[i_x, i_t]



# Fixes for defective mne functions
from collections import defaultdict
import os.path as op
import packaging.version

import numpy as np

from mne.surface import read_surface
from mne.utils import get_subjects_dir, logger, verbose
from mne.label import _get_annot_fname, _n_colors, _write_annot
import nibabel

if packaging.version.parse(nibabel.__version__) == packaging.version.parse('2.3.0'):
    raise ImportError(f"The installed version of nibabel ({nibabel.__version__}) is defective (see https://github.com/nipy/nibabel/issues/649). Please install a newer version.")


@verbose
def write_labels_to_annot(labels, subject=None, parc=None, overwrite=False,
                          subjects_dir=None, annot_fname=None,
                          colormap='hsv', hemi='both', verbose=None):
    """Create a FreeSurfer annotation from a list of labels

    FIX: always write both hemispheres

    Parameters
    ----------
    labels : list with instances of mne.Label
        The labels to create a parcellation from.
    subject : str | None
        The subject for which to write the parcellation for.
    parc : str | None
        The parcellation name to use.
    overwrite : bool
        Overwrite files if they already exist.
    subjects_dir : string, or None
        Path to SUBJECTS_DIR if it is not set in the environment.
    annot_fname : str | None
        Filename of the .annot file. If not None, only this file is written
        and 'parc' and 'subject' are ignored.
    colormap : str
        Colormap to use to generate label colors for labels that do not
        have a color specified.
    hemi : 'both' | 'lh' | 'rh'
        The hemisphere(s) for which to write *.annot files (only applies if
        annot_fname is not specified; default is 'both').
    verbose : bool, str, int, or None
        If not None, override default verbose level (see mne.verbose).

    Notes
    -----
    Vertices that are not covered by any of the labels are assigned to a label
    named "unknown".
    """
    subjects_dir = get_subjects_dir(subjects_dir)

    # get the .annot filenames and hemispheres
    annot_fname, hemis = _get_annot_fname(annot_fname, subject, hemi, parc,
                                          subjects_dir)

    if not overwrite:
        for fname in annot_fname:
            if op.exists(fname):
                raise ValueError('File %s exists. Use "overwrite=True" to '
                                 'overwrite it' % fname)

    # prepare container for data to save:
    to_save = []
    # keep track of issues found in the labels
    duplicate_colors = []
    invalid_colors = []
    overlap = []
    no_color = (-1, -1, -1, -1)
    no_color_rgb = (-1, -1, -1)
    for hemi, fname in zip(hemis, annot_fname):
        hemi_labels = [label for label in labels if label.hemi == hemi]
        n_hemi_labels = len(hemi_labels)

        if n_hemi_labels == 0:
            ctab = np.empty((0, 4), dtype=np.int32)
            ctab_rgb = ctab[:, :3]
        else:
            hemi_labels.sort(key=lambda label: label.name)

            # convert colors to 0-255 RGBA tuples
            hemi_colors = [no_color if label.color is None else
                           tuple(int(round(255 * i)) for i in label.color)
                           for label in hemi_labels]
            ctab = np.array(hemi_colors, dtype=np.int32)
            ctab_rgb = ctab[:, :3]

            # make color dict (for annot ID, only R, G and B count)
            labels_by_color = defaultdict(list)
            for label, color in zip(hemi_labels, ctab_rgb):
                labels_by_color[tuple(color)].append(label.name)

            # check label colors
            for color, names in labels_by_color.items():
                if color == no_color_rgb:
                    continue

                if color == (0, 0, 0):
                    # we cannot have an all-zero color, otherw. e.g. tksurfer
                    # refuses to read the parcellation
                    msg = ('At least one label contains a color with, "r=0, '
                           'g=0, b=0" value. Some FreeSurfer tools may fail '
                           'to read the parcellation')
                    logger.warning(msg)

                if any(i > 255 for i in color):
                    msg = ("%s: %s (%s)" % (color, ', '.join(names), hemi))
                    invalid_colors.append(msg)

                if len(names) > 1:
                    msg = "%s: %s (%s)" % (color, ', '.join(names), hemi)
                    duplicate_colors.append(msg)

            # replace None values (labels with unspecified color)
            if labels_by_color[no_color_rgb]:
                default_colors = _n_colors(n_hemi_labels, bytes_=True,
                                           cmap=colormap)
                safe_color_i = 0  # keep track of colors known to be in hemi_colors
                for i in range(n_hemi_labels):
                    if ctab[i, 0] == -1:
                        color = default_colors[i]
                        # make sure to add no duplicate color
                        while np.any(np.all(color[:3] == ctab_rgb, 1)):
                            color = default_colors[safe_color_i]
                            safe_color_i += 1
                        # assign the color
                        ctab[i] = color

        # find number of vertices in surface
        if subject is not None and subjects_dir is not None:
            fpath = op.join(subjects_dir, subject, 'surf', '%s.white' % hemi)
            points, _ = read_surface(fpath)
            n_vertices = len(points)
        else:
            if len(hemi_labels) > 0:
                max_vert = max(np.max(label.vertices) for label in hemi_labels)
                n_vertices = max_vert + 1
            else:
                n_vertices = 1
            msg = ('    Number of vertices in the surface could not be '
                   'verified because the surface file could not be found; '
                   'specify subject and subjects_dir parameters.')
            logger.warning(msg)

        # Create annot and color table array to write
        annot = np.empty(n_vertices, dtype=np.intp)
        annot[:] = -1
        # create the annotation ids from the colors
        annot_id_coding = np.array((1, 2 ** 8, 2 ** 16))
        annot_ids = list(np.sum(ctab_rgb * annot_id_coding, axis=1))
        for label, annot_id in zip(hemi_labels, annot_ids):
            # make sure the label is not overwriting another label
            if np.any(annot[label.vertices] != -1):
                other_ids = set(annot[label.vertices])
                other_ids.discard(-1)
                other_indices = (annot_ids.index(i) for i in other_ids)
                other_names = (hemi_labels[i].name for i in other_indices)
                other_repr = ', '.join(other_names)
                msg = "%s: %s overlaps %s" % (hemi, label.name, other_repr)
                overlap.append(msg)

            annot[label.vertices] = annot_id

        hemi_names = [label.name for label in hemi_labels]

        # Assign unlabeled vertices to an "unknown" label
        unlabeled = (annot == -1)
        if np.any(unlabeled):
            msg = ("Assigning %i unlabeled vertices to "
                   "'unknown-%s'" % (unlabeled.sum(), hemi))
            logger.info(msg)

            # find an unused color (try shades of gray first)
            for i in range(1, 257):
                if not np.any(np.all((i, i, i) == ctab_rgb, 1)):
                    break
            if i < 256:
                color = (i, i, i, 0)
            else:
                err = ("Need one free shade of gray for 'unknown' label. "
                       "Please modify your label colors, or assign the "
                       "unlabeled vertices to another label.")
                raise ValueError(err)

            # find the id
            annot_id = np.sum(annot_id_coding * color[:3])

            # update data to write
            annot[unlabeled] = annot_id
            ctab = np.vstack((ctab, color))
            hemi_names.append("unknown")

        # convert to FreeSurfer alpha values
        ctab[:, 3] = 255 - ctab[:, 3]

        # remove hemi ending in names
        hemi_names = [name[:-3] if name.endswith(hemi) else name
                      for name in hemi_names]

        to_save.append((fname, annot, ctab, hemi_names))

    issues = []
    if duplicate_colors:
        msg = ("Some labels have the same color values (all labels in one "
               "hemisphere must have a unique color):")
        duplicate_colors.insert(0, msg)
        issues.append('\n'.join(duplicate_colors))
    if invalid_colors:
        msg = ("Some labels have invalid color values (all colors should be "
               "RGBA tuples with values between 0 and 1)")
        invalid_colors.insert(0, msg)
        issues.append('\n'.join(invalid_colors))
    if overlap:
        msg = ("Some labels occupy vertices that are also occupied by one or "
               "more other labels. Each vertex can only be occupied by a "
               "single label in *.annot files.")
        overlap.insert(0, msg)
        issues.append('\n'.join(overlap))

    if issues:
        raise ValueError('\n\n'.join(issues))

    # write it
    for fname, annot, ctab, hemi_names in to_save:
        logger.info('   writing %d labels to %s' % (len(hemi_names), fname))
        _write_annot(fname, annot, ctab, hemi_names)

    logger.info('[done]')



# dataset: mne_sample
"""
This example demonstrates how to load and plot MEG data

It uses the mne sample dataset which included presentaton of simple (A)uditory
and (V)isual stimuli to the (L)eft and (R)ight ear/visual hemifield, as well as
presentation of smiley faces and button presses by the subject.

All the lines starting with "print" are there just to display information about
the data and could be left out for a pure analysis script.

"""
import os
import mne
from eelbrain import *

# find the path for the mne sample data file
datapath = mne.datasets.sample.data_path()
raw_path = os.path.join(datapath, 'MEG', 'sample',
                        'sample_audvis_filt-0-40_raw.fif')

# Load the events from the samepl data file
ds = load.mne.events(raw_path)

# print the first 10 events to check what we loaded
print(ds[:10])
# check how many events of which trigger we have
print(table.frequencies('trigger', data=ds))

# retrieve the trigger variable form the dataset for easier access
trigger = ds['trigger']
# use the trigger variable to add more meaningful labels to the dataset
condition_labels = {1:'LA', 2:'RA', 3:'LV', 4:'RV', 5:'smiley', 32:'button'}
ds['condition'] = Factor(trigger, labels=condition_labels)
side_labels = {1: 'L', 2:'R', 3:'L', 4:'R', 5:'None', 32:'None'}
ds['side'] = Factor(trigger, labels=side_labels)
modality_labels = {1: 'A', 2:'A', 3:'V', 4:'V', 5:'None', 32:'None'}
ds['modality'] = Factor(trigger, labels=modality_labels)

# print the first 10 events with the new labels
print(ds[:10])
# print a nicer table with event frequencies
print(table.frequencies('side', 'modality', data=ds))
# don't print None cells
print(table.frequencies('side', 'modality', data=ds, sub="modality != 'None'"))

# Extract a subset of events for which to load data
ds_sub = ds.sub("modality != 'None'")
# this would be doing the same more explicitly:
##side = ds['side']
##index = side == 'left'
##ds_left = ds.sub(index)

# Load epochs for our selection. Baseline correct from the beginning of the
# epoch to the trigger i.e., t=0). Reject trials with peak to peak values larger
# than 3 pico tesla.
ds_sub = load.mne.add_epochs(ds_sub, -0.1, 0.6, baseline=(None, 0),
                              reject=3e-12, sysname='neuromag306mag')
# check how many events are left
print(table.frequencies('modality', data=ds_sub))

# Plot a butterfly plot with flexible topography of the grand average
plot.TopoButterfly('meg', data=ds_sub)
# plot all conditions separately ('%' is to specify an interaction)
plot.TopoButterfly('meg', 'side % modality', data=ds_sub)

# compare left and right visual stimulation in a cluster t-test (this might
# take a minute -- increase samples to 10000 for a better test)
res = testnd.TTestIndependent('meg', 'side', 'L', 'R', sub="modality == 'V'",
                              data=ds_sub, samples=100, pmin=0.05, tstart=0,
                              tstop=0.4, mintime=0.01)
# show parameters for all clusters
print(res.clusters)
# show parameters for cluster with p <= 0.1 (res.clusters is a Dataset)
print(res.clusters.sub("p <= 0.1"))
# plot the significant clusters
p = plot.TopoArray(res)
p.set_topo_ts(0.1, 0.17, 0.26)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from os.path import join

import mne
from numpy.testing import assert_array_equal
import numpy as np
import pytest

from eelbrain import gui, load, save, set_log_level
from eelbrain.testing import TempDir, gui_test
from eelbrain._wxgui.select_epochs import Document, Model


@gui_test
def test_select_epochs():
    "Test Select-Epochs GUI Document"
    set_log_level('warning', 'mne')

    data_path = mne.datasets.testing.data_path()
    raw_path = join(data_path, 'MEG', 'sample', 'sample_audvis_trunc_raw.fif')
    raw = mne.io.Raw(raw_path, preload=True).pick_types('mag', stim=True)
    ds = load.mne.events(raw, stim_channel='STI 014')
    ds['meg'] = load.mne.epochs(ds, tmax=0.1)
    # 25 cases
    arange = np.arange(25)
    def false_at(index): return np.isin(arange, index, invert=True)

    tempdir = TempDir()
    path = join(tempdir, 'rej.pickle')

    # Test Document
    # =============
    # create a file
    doc = Document(ds, 'meg')
    doc.set_path(path)
    doc.set_case(1, False, 'tag', None)
    doc.set_case(slice(22, 24), False, 'tag', None)
    doc.set_case(2, None, None, ['2'])
    doc.set_bad_channels([1])
    # check modifications
    assert_array_equal(doc.accept, false_at([1, 22, 23]))
    assert doc.tag[1] == 'tag'
    assert doc.interpolate[1] == []
    assert doc.interpolate[2] == ['2']
    assert doc.bad_channels == [1]
    # save
    doc.save()

    # check the file
    ds_ = load.unpickle(path)
    assert doc.epochs.sensor._array_index(ds_.info['bad_channels']) == [1]

    # load the file
    doc = Document(ds, 'meg', path=path)
    # modification checks
    assert_array_equal(doc.accept, false_at([1, 22, 23]))
    assert doc.tag[1] == 'tag'
    assert doc.interpolate[1] == []
    assert doc.interpolate[2] == ['2']
    assert doc.bad_channels == [1]

    # Test Model
    # ==========
    doc = Document(ds, 'meg', path=path)
    model = Model(doc)

    # accept
    assert_array_equal(doc.accept, false_at([1, 22, 23]))
    model.set_case(0, False, None, None)
    assert_array_equal(doc.accept, false_at([0, 1, 22, 23]))
    model.history.undo()
    assert_array_equal(doc.accept, false_at([1, 22, 23]))
    model.history.redo()
    assert_array_equal(doc.accept, false_at([0, 1, 22, 23]))

    # interpolate
    model.toggle_interpolation(2, '2')
    model.toggle_interpolation(2, '3')
    assert doc.interpolate[2] == ['3']
    model.toggle_interpolation(2, '4')
    assert doc.interpolate[2] == ['3', '4']
    model.toggle_interpolation(2, '3')
    assert doc.interpolate[2] == ['4']
    model.toggle_interpolation(3, '3')
    assert doc.interpolate[2] == ['4']
    assert doc.interpolate[3] == ['3']
    model.history.undo()
    model.history.undo()
    assert doc.interpolate[2] == ['3', '4']
    assert doc.interpolate[3] == []
    model.history.redo()
    assert doc.interpolate[2] == ['4']

    # bad channels
    model.set_bad_channels([1])
    model.set_bad_channels([1, 10])
    assert doc.bad_channels == [1, 10]
    model.history.undo()
    assert doc.bad_channels == [1]
    model.history.redo()
    assert doc.bad_channels == [1, 10]

    # reload to reset
    model.load(path)
    assert_array_equal(doc.accept, false_at([1, 22, 23]))
    assert doc.tag[1] == 'tag'
    assert doc.interpolate[1] == []
    assert doc.interpolate[2] == ['2']
    assert doc.bad_channels == [1]

    # load truncated file
    rej_ds = load.unpickle(path)
    save.pickle(rej_ds[:23], path)
    with pytest.raises(IOError):
        model.load(path, answer=False)
    model.load(path, answer=True)
    assert_array_equal(doc.accept, false_at([1, 22]))

    # Test GUI
    # ========
    frame = gui.select_epochs(ds, nplots=9)
    assert not frame.CanBackward()
    assert frame.CanForward()
    frame.OnForward(None)
    frame.SetVLim(1e-12)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-trf_intro:
.. currentmodule:: eelbrain

Introduction to Temporal Response Functions (TRFs)
==================================================

A Temporal Response Function (TRF) is a linear stimulus-response model.
The predicted response is a linear convolution of the stimulus with the TRF.
A simple illustration of this is that
every impulse in the stimulus is associated with a corresponding response,
shaped like the TRF:
"""
from eelbrain import *
import numpy as np

# Construct a 10 s long stimulus
time = UTS(0, 0.01, 1000)
x = NDVar(np.zeros(len(time)), time)
# add a few impulses
x[1] = 1
x[3] = 1
x[5] = 1

# Construct a TRF of length 500 ms
trf_time = UTS(0, 0.01, 50)
trf = gaussian(0.200, 0.050, trf_time) - gaussian(0.300, 0.050, trf_time)

# The response is the convolution of the stimulus with the TRF
y = convolve(trf, x)

plot_args = dict(columns=1, axh=2, w=10, frame='t', legend=False, colors='r')
p = plot.UTS([x, trf, y], ylabel=['Stimulus (x)', 'TRF', 'Response (y)'], **plot_args)

###############################################################################
# Because the convolution is linear:
#
#  - Scaled stimuli cause scaled responses
#  - Overlapping responses add up

x[2] = 0.66
x[5.2] = 1
x[7] = 0.8
x[7.2] = 0.3
x[8.8] = 0.2
x[9] = 0.5
x[9.2] = 0.7

y = convolve(trf, x)
p = plot.UTS([x, trf, y], ylabel=['Stimulus (x)', 'TRF', 'Response (y)'], **plot_args)

###############################################################################
# When the stimulus contains only non-zero elements this works just the same,
# but the TRF shape may not be apparent in the response anymore:

x += np.random.normal(0, 0.1, x.shape)
filter_data(x, 1, 40)
y = convolve(trf, x)
p =  plot.UTS([x, trf, y], ylabel=['Stimulus (x)', 'TRF', 'Response (y)'], **plot_args)

###############################################################################
# Given a stimulus and a response, there are different methods to reconstruct
# the TRF. Eelbrain comes with an implementation of :func:`boosting`, a
# coordinate descent algorithm with early stopping based on cross-validation:

fit = boosting(y, x, 0.000, 0.500, basis=0.050, partitions=2)
plot_args = {**plot_args, 'columns': 3}
p = plot.UTS([trf, fit.h, None], axtitle=["Model TRF", "Reconstructed TRF"], **plot_args)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Statistical Parametric Mapping"""
from itertools import repeat
from typing import Dict, Literal, Sequence, Union

import numpy as np

from .. import _info, fmtxt
from .._data_obj import Dataset, Factor, Var, NDVar, Case, IndexArg, ModelArg, NDVarArg, Parametrization, PermutedParametrization, asmodel, asndvar, assub, combine
from .._exceptions import DimensionMismatchError
from .._utils import deprecate_ds_arg
from . import stats
from .glm import MPTestMapper
from .test import star
from .testnd import TTestOneSample, MultiEffectNDTest, NDPermutationDistribution, permute_order, run_permutation_me


class LMMapper(MPTestMapper):

    def __init__(
            self,
            parametrization: Parametrization,
    ):
        self.parametrization = parametrization
        self.permuted_parametrization = PermutedParametrization(parametrization, g=True)
        self._flat_t_buffer = None

    def preallocate(self, y_shape: Sequence[int]) -> np.ndarray:
        n_columns = len(self.parametrization.column_names)
        shape = (n_columns, *y_shape)
        buffer = np.empty(shape)
        self._flat_t_buffer = buffer.reshape((n_columns, -1))
        return buffer

    def map(
            self,
            y: np.ndarray,  # (n_cases, ...)
            perm: np.ndarray = None,  # (n_cases,) permutation index
    ) -> None:
        if perm is None:
            parametrization = self.parametrization
        else:
            parametrization = self.permuted_parametrization
            parametrization.permute(perm)
        stats.lm_t(y, parametrization, out_t=self._flat_t_buffer)


class LM(MultiEffectNDTest):
    """Fixed effects linear model

    Parameters
    ----------
    y
        Dependent variable.
    x
        Model to fit.
    sub
        Only use part of the data.
    data
        Optional Dataset providing data for y/model.
    coding
        Model parametrization (default is dummy coding). Vars are centered for
        effect coding (but not for dummy coding).
    subject
        Optional information used by :class:`LMGroup`; if subject is a column in
        ``ds`` it will be extracted automatically.
    samples
        Number of samples for permutation test (default 10,000).
    pmin
        Threshold for forming clusters:  use a t-value equivalent to an
        uncorrected p-value.
    tmin
        Threshold for forming clusters as t-value.
    tfce
        Use threshold-free cluster enhancement. Use a scalar to specify the
        step of TFCE levels (for ``tfce is True``, 0.1 is used).
    tstart
        Start of the time window for the permutation test (default is the
        beginning of ``y``).
    tstop
        Stop of the time window for the permutation test (default is the
        end of ``y``).
    force_permutation
        Conduct permutations regardless of whether there are any clusters.
    mintime : scalar
        Minimum duration for clusters (in seconds).
    minsource : int
        Minimum number of sources per cluster.

    See Also
    --------
    LMGroup

    Examples
    --------
    See :ref:`exa-two-stage` example.

    Notes
    -----
    By default, this model generates a permutation distribution to correct for
    multiple comparisons. This is not needed for a two-stage model, where
    correction occurs at the group level. When fitting two-stage models, set
    ``samples=0`` to skip this and save time.

    This class stores a shallow copy of ``y.info`` (for predicting).
    """
    _state_specific = ('model', 'coding', '_coeffs_flat', '_se_flat', 'subject_variables', '_parametrization', '_y_info')
    # _statistic = 't'  # would be consistent with testnd but require turning LM.t() into an attr

    @deprecate_ds_arg
    def __init__(
            self,
            y: NDVarArg,
            x: ModelArg,
            sub: IndexArg = None,
            data: Dataset = None,
            coding: Literal['dummy', 'effect'] = 'dummy',
            subject: str = None,
            samples: int = 10000,
            pmin: float = None,
            tmin: float = None,
            tfce: Union[float, bool] = False,
            tstart: float = None,
            tstop: float = None,
            force_permutation: bool = False,
            **criteria,
    ):
        sub_arg = sub
        sub, n_cases = assub(sub, data, return_n=True)
        y, n_cases = asndvar(y, sub, data, n_cases, return_n=True)
        model = asmodel(x, sub, data, n_cases)
        parametrization = model._parametrize(coding)
        ß_maps, se_maps, t_maps = stats.lm_t(y.x, parametrization)
        # find variables to keep
        variables = {}
        if data is not None:
            for key, item in data.items():
                if isinstance(item, Factor):
                    if sub is not None:
                        item = item[sub]
                    if len(item.cells) == 1:
                        variables[key] = item.cells[0]
        # subject
        if subject is not None:
            if not isinstance(subject, str):
                raise TypeError(f"{subject=}: needs to be string or None")
            variables['subject'] = subject

        # Cluster-based tests
        n_effects = len(parametrization.column_names)
        n_threshold_params = sum((pmin is not None, tmin is not None, bool(tfce)))
        if n_threshold_params == 0 and not samples:
            cdists = None
        elif n_threshold_params > 1:
            raise ValueError("Only one of pmin, tmin and tfce can be specified")
        else:
            if pmin is not None:
                df = len(y) - model.df
                thresholds = tuple(repeat(stats.ttest_t(pmin, df), n_effects))
            elif tmin is not None:
                thresholds = tuple(repeat(abs(tmin), n_effects))
            else:
                thresholds = tuple(repeat(None, n_effects))

            cdists = [NDPermutationDistribution(y, samples, thresh, tfce, 0, 't', name, tstart, tstop, criteria, None, force_permutation) for name, thresh in zip(parametrization.column_names, thresholds)]

            # Find clusters in the actual data
            do_permutation = 0
            for cdist, t_map in zip(cdists, t_maps):
                cdist.add_original(t_map)
                do_permutation += cdist.do_permutation
            # Generate null distribution
            if do_permutation:
                iterator = permute_order(len(y), samples)
                run_permutation_me(LMMapper(parametrization), cdists, iterator)

        x_desc = x if isinstance(x, str) else model.name  # TODO: x.name should use * when appropriate
        MultiEffectNDTest.__init__(self, x_desc, parametrization.column_names, y, None, sub_arg, samples, tfce, pmin, cdists, tstart, tstop)
        self.coding = coding
        self._coeffs_flat = ß_maps.reshape((len(ß_maps), -1))
        self._se_flat = se_maps.reshape((len(se_maps), -1))
        self.model = model
        self._parametrization = parametrization
        self._y_info = y.info.copy()
        self.subject_variables = variables
        self._expand_state()

    def _expand_state(self):
        self.subject = self.subject_variables.get('subject', None)
        self._shape = tuple(map(len, self._dims))
        self.column_names = self._parametrization.column_names
        self.n_cases = self.model.df_total
        self.dims = self._dims
        MultiEffectNDTest._expand_state(self)

    def _name(self):
        y = f'{self.y} ' if self.y else ''
        return f"LM: {y}~ {self.model.name}, {self.subject}>"

    def __setstate__(self, state):
        # backwards compatibility
        if 'dims' in state:
            state['_dims'] = state.pop('dims')
            state['_coeffs_flat'] = state.pop('coeffs')
            state['_se_flat'] = state.pop('se')
            if 'subject' in state:
                state['subject_variables'] = {'subject': state.pop('subject')}
            if 'p' in state:
                state['_parametrization'] = state.pop('p')
            else:
                state['_parametrization'] = state['model']._parametrize(state['coding'])
        MultiEffectNDTest.__setstate__(self, state)

    def _coefficient(self, term):
        """Regression coefficient for a given term"""
        index = self._index(term)
        return self._coeffs_flat[index].reshape((1,) + self._shape)

    def _default_plot_obj(self):
        if self.samples:
            return [self.masked_parameter_map(e) for e in self.effects]
        else:
            return [self.t(term) for term in self.column_names]

    def _index(self, term: str) -> int:
        if term in self.column_names:
            return self.column_names.index(term)
        elif term in self._parametrization.terms:
            index = self._parametrization.terms[term]
            if index.stop - index.start > 1:
                raise NotImplementedError("Term has more than one column")
            return index.start
        else:
            raise KeyError(f"{term=}")

    def coefficient(self, term):
        ":class:`NDVar` with regression coefficient for a given term (or ``'intercept'``)"
        return NDVar(self._coefficient(term)[0], self.dims, term, self._y_info)

    def predict(
            self,
            values: Union[Sequence[float], Dict[str, float]],
            name: str = None,
    ) -> NDVar:
        """Predict ``y`` based on given values of ``x``

        Parameters
        ----------
        values
            Give as list of values, in the same order as ``x``, or as dictionary
            mapping predictor names to values (missing predictors are
            substituted with the mean from the original data).
        name
            Name the resulting :class:`NDVar`.
        """
        names = [e.name for e in self._parametrization.model.effects]
        if isinstance(values, dict):
            if unknown := set(values).difference(names):
                raise ValueError(f"{values}: Unknown terms ({', '.join(unknown)})")
            x_in = []
            for e in self._parametrization.model.effects:
                if isinstance(e, Var):
                    if e.name in values:
                        v_i = values[e.name]
                        if self._parametrization.method == 'effect':
                            v_i -= e.mean()
                    elif self._parametrization.method == 'effect':
                        v_i = 0
                    else:
                        v_i = e.mean()
                    x_in.append(v_i)
                else:
                    raise NotImplementedError("Predict for categorial models")
        else:
            x_in = list(values)
            if (l1 := len(x_in)) != (l2 := len(self._parametrization.x.shape(1) - 1)):
                raise ValueError(f"{values}: wrong length (got {l1}, need {l2})")
            if self._parametrization.method == 'effect':
                x_in = [v / e.mean() for e, v in zip(self._parametrization.model.effects, x_in)]
        # Add intercept
        x = np.append(1, x_in)
        y = x.dot(self._coeffs_flat)
        return NDVar(y.reshape(self._shape), self.dims, name, self._y_info)

    def t(self, term: str) -> NDVar:
        ":class:`NDVar` with t-values for a given term (or ``'intercept'``)."
        index = self._index(term)
        se = self._se_flat[index]
        flat_index = se == 0.
        any_flat = np.any(flat_index)
        if any_flat:
            se = se.copy()
            se[flat_index] = 1.
        t = self._coeffs_flat[index] / se
        if any_flat:
            t[np.logical_and(flat_index, t == 0)] = 0.
            t[np.logical_and(flat_index, t != 0)] *= np.inf
        info = _info.for_stat_map('t')
        info['term'] = term
        return NDVar(t.reshape(self._shape), self.dims, term, info)

    def _n_columns(self):
        return {term: s.stop - s.start for term, s in self._parametrization.terms.items()}


class LMGroup:
    """Group level analysis for linear model :class:`LM` objects

    Parameters
    ----------
    lms : sequence of LM
        A separate :class:`LM` object for each subject.

    Attributes
    ----------
    column_names : [str]
        Names of the linear model columns.
    column_keys : [str]
        Corresponding dataset keys (with invalid characters replaced).
    tests : None | {str: TTestRelated}
        Tests computed with :meth:`compute_column_ttests`.
    samples : None | int
        Number of samples used to compute tests in :attr:`tests`.

    See Also
    --------
    LM

    Examples
    --------
    See :ref:`exa-two-stage` example.
    """
    def __init__(self, lms):
        # check lms
        lm0 = lms[0]
        n_columns_by_term = lm0._n_columns()
        for lm in lms[1:]:
            if lm.dims != lm0.dims:
                raise DimensionMismatchError("LMs have incompatible dimensions")
            elif lm._n_columns() != n_columns_by_term:
                raise ValueError(f"Model for {lm0.subject} and {lm.subject} don't match")
            elif lm.coding != lm0.coding:
                raise ValueError("Models have incompatible coding")

        # make sure to have a unique subject label for each lm
        name_i = 0
        subjects = [lm.subject for lm in lms]
        str_names = tuple(filter(None, subjects))
        if len(set(str_names)) < len(str_names):
            raise ValueError(f"Duplicate subject names: {', '.join(map(repr, str_names))}")
        new_name = 'S000'
        for i in range(len(subjects)):
            if not subjects[i]:
                while new_name in subjects:
                    name_i += 1
                    new_name = 'S%03i' % name_i
                subjects[i] = new_name

        self._lms = lms
        self._subjects = tuple(subjects)
        self.tests = None
        self._init_secondary()

    def __setstate__(self, state):
        self._lms = state['lms']
        self._subjects = state['subjects']
        self.tests = state.get('tests')
        self._init_secondary()

    def _init_secondary(self):
        lm = self._lms[0]
        self.dims = lm.dims
        self.coding = lm.coding
        self.column_names = lm.column_names
        if self.tests is None:
            self.samples = None
        else:
            self.samples = self.tests[self.column_names[0]].samples
        # subject variables
        self.subject_variables = Dataset()
        self.subject_variables['subject'] = Factor(self._subjects, random=True)
        keys = set().union(*(lm.subject_variables for lm in self._lms))
        for key in keys:
            values = [lm.subject_variables.get(key, '') for lm in self._lms]
            self.subject_variables[key] = Factor(values)
        self.column_keys = [Dataset.as_key(name) for name in self.column_names]

    def __getstate__(self):
        return {'lms': self._lms, 'tests': self.tests, 'subjects': self._subjects}

    def __repr__(self):
        lm = self._lms[0]
        return f"<LMGroup {lm.y!r}, {lm.x!r}, n={len(self._lms)}>"

    def coefficients(self, term):
        "Coefficients for one term as :class:`NDVar`"
        x = np.concatenate([lm._coefficient(term) for lm in self._lms])
        return NDVar(x, (Case,) + self.dims, name=term)

    def coefficients_dataset(self, terms=None, long=False):
        """Regression coefficients in a :class:`Dataset`

        By default, each regression coefficient is assigned as separate column.
        With ``long=True``, a long form table is produced, in which the
        coefficients for different terms are stacked vertically and the ``term``
        :class:`Factor` specifies which term the coefficients correspond to.

        Parameters
        ----------
        terms : str | sequence of str
            Terms for which to retrieve coefficients (default is all terms).
        long : bool
            Produce a table in long form.

        Returns
        -------
        ds : Dataset
            If the dataset is in the wide form (default), each model coefficient
            is assigned as :class:`NDVar` with a key corresponding to the
            ``.column_keys`` attribute.
            If the dataset is in the long form (with ``long=True``), it has the
            following entries:

             - ``coeff``: the coefficients in an :class:`NDVar`
             - ``term``: a :class:`Factor` with the name of the term

            In addition, all available subject-variables are assigned.
        """
        if isinstance(terms, str):
            terms = (terms,)
        elif terms is None:
            terms = self.column_names

        if long:
            ds = Dataset({
                'coeff': combine([self.coefficients(term) for term in terms]),
                'term': Factor(terms, repeat=len(self._lms)),
                **self.subject_variables.tile(len(terms)),
            })
        else:
            ds = Dataset({Dataset.as_key(name): self.coefficients(name) for name in terms})
            ds.update(self.subject_variables)
        return ds

    def column_ttest(self, term, return_data=False, popmean=0, *args, **kwargs):
        """One-sample t-test on a single model column

        Parameters
        ----------
        term : str
            Name of the term to test.
        return_data : bool
            Return the individual subjects' coefficients along with test
            results.
        popmean : scalar
            Value to compare y against (default is 0).
        tail : 0 | 1 | -1
            Which tail of the t-distribution to consider:
            0: both (two-tailed);
            1: upper tail (one-tailed);
            -1: lower tail (one-tailed).
        samples : None | int
            Number of samples for permutation cluster test. For None, no
            clusters are formed. Use 0 to compute clusters without performing
            any permutations.
        pmin : None | scalar (0 < pmin < 1)
            Threshold for forming clusters:  use a t-value equivalent to an
            uncorrected p-value.
        tmin : None | scalar
            Threshold for forming clusters.
        tfce : bool
            Use threshold-free cluster enhancement (Smith & Nichols, 2009).
            Default is False.
        tstart, tstop : scalar
            Restrict time window for permutation cluster test.
        mintime : scalar
            Minimum duration for clusters (in seconds).
        minsource : int
            Minimum number of sources per cluster.

        Returns
        -------
        result : TTestOneSample
            T-test result.
        data : Dataset (only with ``return_data=True``)
            Dataset with subjects' coefficients.

        Notes
        -----
        Performs a one-sample t-test on coefficient estimates from all subjects
        to test the hypothesis that the coefficient is different from popmean
        in the population.
        """
        coeff = self.coefficients(term)
        res = TTestOneSample(coeff, popmean, None, None, None, *args, **kwargs)
        if return_data:
            ds = Dataset({'coeff': coeff})
            ds.update(self.subject_variables)
            ds['n'] = Var([lm.n_cases for lm in self._lms])
            return res, ds
        else:
            return res

    def design(self, subject=None):
        "Table with the design matrix"
        if subject is None:
            lm = self._lms[0]
            subject = self._subjects[0]
        else:
            for lm, lm_subject in zip(self._lms, self._subjects):
                if lm_subject == subject:
                    break
            else:
                raise ValueError(f"subject={subject!r}")

        table = lm.model.as_table(lm.coding)
        table.caption(f"Design matrix for {subject}")
        return table

    def compute_column_ttests(self, *args, **kwargs):
        """Compute all tests and store them in :attr:`self.tests`

        Parameters like :meth:`.column_ttest`, starting with ``popmean``.
        """
        self.tests = {}
        for term in self.column_names:
            self.tests[term] = self.column_ttest(term, False, *args, **kwargs)
        self.samples = self.tests[self.column_names[0]].samples

    def info_list(self):
        info = fmtxt.List("LMGroup info")
        for effect in self.column_names:
            res = self.tests[effect]
            info.add_sublist(effect, [res.info_list()])
        return info

    def table(self, title=None, caption=None):
        """Table listing all terms and corresponding smallest p-values

        Parameters
        ----------
        title : text
            Title for the table.
        caption : text
            Caption for the table.

        Returns
        -------
        table : eelbrain.fmtxt.Table
            ANOVA table.
        """
        if not self.tests:
            raise RuntimeError("Need to precompute tests with .compute_column_ttests()")
        table = fmtxt.Table('lrrl', title=title, caption=caption)
        table.cells('Term', 't_max', 'p', 'sig')
        table.midrule()
        for term, res in self.tests.items():
            table.cell(term)
            table.cell(fmtxt.stat(res.t.extrema()))
            pmin = res.p.min()
            table.cell(fmtxt.p(pmin))
            table.cell(star(pmin))
        return table


# for backwards compatibility
RandomLM = LMGroup



# skip test: data unavailable
"""Generate group level result resports for the Mouse experiment

This script generates HTML reports based on the experiment defined in
``mouse.py``.
"""
from mouse import e


e.set(epoch='target', inv='fixed-3-dSPM')
e.make_report('surprise', mask='frontotemporal-lh', pmin=0.05, baseline=False, tstart=0.3, tstop=0.5, raw='1-40')
e.make_report('surprise', mask='frontotemporal-lh', pmin=0.05, baseline=False, tstart=0.3, tstop=0.5, raw='fastica1-40')



from eelbrain import fmtxt

table = fmtxt.Table('lll')
table.cell("Animal", r"\textbf")
table.cell("Outside", r"\textbf")
table.cell("Inside", r"\textbf")
table.midrule()
table.cell("Duck")
table.cell("Feathers")
table.cell("Duck Meat")
table.cell("Dog")
table.cell("Fur")
table.cell("Hotdog Meat")

# print the string representation
print(table)



import numpy as np
from eelbrain._stats import opt, stats

np.random.seed(0)

n_cases = 100
n_tests = 1000
n_betas = 3
n_fs = 2

# lm
y = np.random.normal(0, 1, (n_cases, n_tests))
x = np.random.normal(0, 1, (n_cases, n_betas))
xsinv = np.random.normal(0, 1, (n_betas, n_cases))
betas = np.empty((n_tests, n_betas))
ss = np.empty(n_tests)
perm = np.arange(n_cases)
np.random.shuffle(perm)

# bigger arrays
n_testsl = 10**6
yl = np.random.normal(0, 1, (n_cases, n_testsl))
betasl = np.empty((n_testsl, n_betas))
ssl = np.empty(n_testsl)


# ANOVA
f_map = np.empty((n_fs, n_tests))
effects = np.array([[0, 2], [2, 1]], dtype=np.int16)
df_res = n_cases - n_betas - 1

print("n_cases=%i; n_tests=%i; n_betas=%i" % (n_cases, n_tests, n_betas))
print("timeit opt.lm_betas(y, x, xsinv, betas)")
print("timeit -n1000 opt.lm_res_ss(y, x, xsinv, ss)")
print("timeit -n1000 opt._anova_fmaps(y, x, xsinv, f_map, effects, df_res)")
print("timeit opt._ss(y, ss)")



"""Test mne interaction"""
import os
from pathlib import Path
import warnings

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import mne
from nibabel.freesurfer import read_annot
import pytest

from eelbrain import (
    datasets, load, table, testnd,
    Dataset, Factor,
    concatenate, labels_from_clusters, morph_source_space, set_parc, xhemi)
from eelbrain._data_obj import SourceSpace, asndvar, _matrix_graph
from eelbrain._mne import shift_mne_epoch_trigger, combination_label
from eelbrain.testing import requires_mne_sample_data, requires_mne_testing_data
from eelbrain.tests.test_data import assert_dataobj_equal


def assert_label_equal(l0, l1, comment=True, color=True):
    if isinstance(l0, mne.BiHemiLabel):
        assert isinstance(l1, mne.BiHemiLabel)
        assert_label_equal(l0.lh, l1.lh, comment, color)
        assert_label_equal(l0.rh, l1.rh, comment, color)
    else:
        if comment:
            assert l0.comment == l1.comment
        if color:
            assert l0.color == l1.color
        assert l0.hemi == l1.hemi
        assert l0.subject == l1.subject
        assert_array_equal(l0.vertices, l1.vertices)
        assert_allclose(l0.pos, l1.pos)
        assert_allclose(l0.values, l1.values)


@requires_mne_sample_data
def test_source_estimate():
    "Test SourceSpace dimension"
    mne.set_log_level('warning')
    # load data
    ds = datasets.get_mne_sample(src='ico')
    dsa = ds.aggregate('side')
    src = ds['src']
    source = src.source

    # test auto-conversion
    asndvar('epochs', data=ds)
    asndvar('epochs', data=dsa)
    asndvar(dsa['epochs'][0])

    # source space clustering
    res = testnd.TTestIndependent('src', 'side', data=ds, samples=0, pmin=0.05, tstart=0.05, mintime=0.02, minsource=10)
    assert res.clusters.n_cases == 52

    # test disconnecting parc
    orig_conn = set(map(tuple, source.adjacency()))
    disc_conn = set(map(tuple, source.adjacency(True)))
    assert len(disc_conn) < len(orig_conn)
    for pair in orig_conn:
        s, d = pair
        if pair in disc_conn:
            assert source.parc[s] == source.parc[d]
        else:
            assert source.parc[s] != source.parc[d]

    # threshold-based test with parc
    srcl = src.sub(source='lh')
    res = testnd.TTestIndependent(srcl, 'side', data=ds, samples=10, pmin=0.05, tstart=0.05, mintime=0.02, minsource=10, parc='source')
    assert res._cdist.dist.shape[1] == len(srcl.source.parc.cells)
    label = 'superiortemporal-lh'
    c_all = res.find_clusters(maps=True)
    c_label = res.find_clusters(maps=True, source=label)
    assert_array_equal(c_label['location'], label)
    for case in c_label.itercases():
        id_ = case['id']
        idx = c_all['id'].index(id_)[0]
        assert case['v'] == c_all[idx, 'v']
        assert case['tstart'] == c_all[idx, 'tstart']
        assert case['tstop'] == c_all[idx, 'tstop']
        assert case['p'] <= c_all[idx, 'p']
        assert_dataobj_equal(case['cluster'], c_all[idx, 'cluster'].sub(source=label))

    # threshold-free test with parc
    res = testnd.TTestIndependent(srcl, 'side', data=ds, samples=10, tstart=0.05, parc='source')
    cl = res.find_clusters(0.05)
    assert cl.eval("p.min()") == res.p.min()
    mp = res.masked_parameter_map()
    assert mp.min() == res.t.min()
    assert mp.max() == res.t.max(res.p <= 0.05)
    assert mp.max() == pytest.approx(-4.95817732)

    # indexing source space
    s_sub = src.sub(source='fusiform-lh')
    idx = source.index_for_label('fusiform-lh')
    s_idx = src[idx]
    assert_dataobj_equal(s_sub, s_idx)

    # concatenate
    src_reconc = concatenate((src.sub(source='lh'), src.sub(source='rh')), 'source')
    assert_dataobj_equal(src_reconc, src)
    # label
    src_sm = src.sub(source=['superiortemporal-lh', 'middletemporal-lh'])
    src_s = src.sub(source='superiortemporal-lh')
    src_m = src.sub(source='middletemporal-lh')
    src_conc = concatenate([src_s, src_m], 'source')
    assert_dataobj_equal(src_conc, src_sm)
    assert_array_equal(src_conc.source.parc, src_sm.source.parc)

    # melt NDVar
    dsa['src_sub'] = dsa['src'].sub(source=['lateraloccipital-lh', 'lateraloccipital-rh'])
    mdsa = table.melt_ndvar('src_sub', 'source', ds=dsa)
    assert_array_equal(mdsa['hemi'], Factor(['lh'] * 88 * 2 + ['rh'] * 87 * 2))


@requires_mne_sample_data
def test_dataobjects():
    "Test handing MNE-objects as data-objects"
    shift = np.array([0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, -0.1])
    epochs = datasets.get_mne_epochs()
    ds = Dataset({'a': Factor('ab', repeat=8), 'epochs': epochs})
    ds['ets'] = shift_mne_epoch_trigger(epochs, shift, min(shift), max(shift))

    # ds operations
    ds.sub("a == 'a'")
    ds.aggregate('a')

    # asndvar
    asndvar(ds['epochs'])
    asndvar(ds['ets'])

    # adjacency
    ds = datasets.get_mne_sample(sub=[0], sns=True)
    sensor = ds['meg'].sensor
    c = sensor.adjacency()
    assert_array_equal(c[:, 0] < c[:, 1], True)
    assert c.max() == len(sensor) - 1


@requires_mne_sample_data
def test_epoch_trigger_shift():
    "Test the shift_mne_epoch_trigger() function"
    epochs = datasets.get_mne_sample(sns=True, sub="[1,2,3]")['epochs']
    n_lost_start = np.sum(epochs.times < epochs.tmin + 0.05)
    n_lost_end = np.sum(epochs.times > epochs.tmax - 0.05)
    data = epochs.get_data()

    # don't shift
    epochs_s = shift_mne_epoch_trigger(epochs, [0, 0, 0])
    assert_array_equal(epochs_s.get_data(), data)

    epochs_s = shift_mne_epoch_trigger(epochs, [-0.05, 0., 0.05])
    data_s = epochs_s.get_data()
    assert_array_equal(data_s[0], data[0, :, : -(n_lost_end + n_lost_start)])
    assert_array_equal(data_s[1], data[1, :, n_lost_start: -n_lost_end])
    assert_array_equal(data_s[2], data[2, :, n_lost_end + n_lost_start:])
    assert_allclose(epochs_s.times, epochs.times[n_lost_start: -n_lost_end],
                    rtol=1e-1, atol=1e-3)  # ms accuracy

    epochs_s = shift_mne_epoch_trigger(epochs, [0.05, 0., 0.05])
    data_s = epochs_s.get_data()
    assert_array_equal(data_s[0], data[0, :, n_lost_end:])
    assert_array_equal(data_s[1], data[1, :, :-n_lost_end])
    assert_array_equal(data_s[2], data[2, :, n_lost_end:])
    assert_allclose(epochs_s.times, epochs.times[:-n_lost_end],
                    rtol=1e-1, atol=1e-3)  # ms accuracy


@requires_mne_testing_data
def test_combination_label():
    "Test combination label creation"
    data_dir = mne.datasets.testing.data_path(download=False)
    subjects_dir = os.path.join(data_dir, 'subjects')

    labels = {l.name: l for l in mne.read_labels_from_annot('fsaverage', subjects_dir=subjects_dir)}

    # standard
    l = combination_label('temporal', "superiortemporal + middletemporal + inferiortemporal", labels, subjects_dir)
    lh = labels['superiortemporal-lh'] + labels['middletemporal-lh'] + labels['inferiortemporal-lh']
    lh.name = 'temporal-lh'
    rh = labels['superiortemporal-rh'] + labels['middletemporal-rh'] + labels['inferiortemporal-rh']
    rh.name = 'temporal-rh'
    assert len(l) == 2
    assert_label_equal(l[0], lh)
    assert_label_equal(l[1], rh)

    # only rh
    l = combination_label('temporal-rh', "superiortemporal + middletemporal + inferiortemporal",
                          labels, subjects_dir)
    assert len(l) == 1
    assert l[0].name == 'temporal-rh'
    assert_array_equal(l[0].vertices, rh.vertices)

    # with split_label
    l2 = combination_label('temporal-rh', "superiortemporal + middletemporal +"
                                          "split(inferiortemporal, 2)[0] +"
                                          "split(inferiortemporal, 2)[1]",
                           labels, subjects_dir)
    assert_label_equal(l2[0], l[0], comment=False, color=False)

    # names with .
    labels = {l.name: l for l in
              mne.read_labels_from_annot('fsaverage', 'PALS_B12_Lobes', subjects_dir=subjects_dir)}
    l = combination_label('frontal-lh', "LOBE.FRONTAL", labels, subjects_dir)[0]
    assert_array_equal(l.vertices, labels['LOBE.FRONTAL-lh'].vertices)


@requires_mne_testing_data
def test_morphing():
    # Test data
    data_dir = Path(mne.datasets.testing.data_path(download=False))
    subjects_dir = data_dir / 'subjects'
    stc = datasets.get_mne_stc()
    ndvar = load.mne.stc_ndvar(stc, 'sample', 'ico-5', subjects_dir, 'dSPM', name='src')
    labels_sample = {label.name: label for label in mne.read_labels_from_annot('sample', 'aparc', subjects_dir=subjects_dir)}
    labels_fsa = {label.name: label for label in mne.read_labels_from_annot('fsaverage', 'aparc', subjects_dir=subjects_dir)}
    morph = mne.compute_source_morph(stc, 'sample', 'fsaverage', subjects_dir)
    stc_fsa = morph.apply(stc)
    stc_fsa_ndvar = load.mne.stc_ndvar(stc_fsa, 'fsaverage', 'ico-5', subjects_dir, 'dSPM', False, 'src', parc=None)

    # sample to fsaverage
    # -------------------
    # Using only subject
    ndvar_fsa = morph_source_space(ndvar, 'fsaverage')
    assert_dataobj_equal(ndvar_fsa, stc_fsa_ndvar, decimal=6)
    # Using pre-computed morph-matrix
    ndvar_fsa = morph_source_space(ndvar, 'fsaverage', morph=morph)
    assert_dataobj_equal(ndvar_fsa, stc_fsa_ndvar, decimal=6)
    # Subset of STC using morph-matrix (equivalent to using stc.expand())
    for hemi in ['lh', 'rh', 'bi']:
        if hemi == 'bi':
            label_sample = labels_sample['transversetemporal-lh'] + labels_sample['transversetemporal-rh']
            label_fsa = labels_fsa['transversetemporal-lh'] + labels_fsa['transversetemporal-rh']
        else:
            label_sample = labels_sample[f'transversetemporal-{hemi}']
            label_fsa = labels_fsa[f'transversetemporal-{hemi}']
        stc_label = stc.in_label(label_sample)
        ndvar_label = load.mne.stc_ndvar(stc_label, 'sample', 'ico-5', subjects_dir, 'dSPM', name='src')
        stc_label_fsa = morph.apply(stc_label.copy().expand(stc.vertices)).in_label(label_fsa)
        stc_label_fsa_ndvar = load.mne.stc_ndvar(stc_label_fsa, 'fsaverage', 'ico-5', subjects_dir, 'dSPM', False, 'src', parc=None)
        ndvar_label_fsa = morph_source_space(ndvar_label, 'fsaverage', morph=morph)
        assert_dataobj_equal(ndvar_label_fsa, stc_label_fsa_ndvar, decimal=6)
        # Single time point
        ndvar_label_fsa = morph_source_space(ndvar_label.sub(time=0), 'fsaverage', morph=morph)
        assert_dataobj_equal(ndvar_label_fsa, stc_label_fsa_ndvar.sub(time=0), decimal=6)
        # Subset of STC (computing morph-matrix for subset)
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'\d+/\d+ vertices not included in smoothing', module='mne')
            label_morph = mne.compute_source_morph(stc_label, 'sample', 'fsaverage', subjects_dir)
        stc_label_fsa = label_morph.apply(stc_label).in_label(label_fsa)
        stc_label_fsa_ndvar = load.mne.stc_ndvar(stc_label_fsa, 'fsaverage', 'ico-5', subjects_dir, 'dSPM', False, 'src', parc=None)
        ndvar_label_fsa = morph_source_space(ndvar_label, 'fsaverage')
        assert_dataobj_equal(ndvar_label_fsa, stc_label_fsa_ndvar, decimal=6)

    # Scaled brain to fsaverage
    # -------------------------
    y_scaled = datasets.get_mne_stc(True, subject='fsaverage_scaled')
    y_scaled_m = morph_source_space(y_scaled, 'fsaverage')
    assert y_scaled_m.source.subject == 'fsaverage'
    assert_array_equal(y_scaled_m.x, y_scaled.x)

    # masked data
    y_sub = y_scaled.sub(source='superiortemporal-lh')
    y_sub_m = morph_source_space(y_sub, 'fsaverage')
    assert y_sub_m.source.subject == 'fsaverage'
    assert_array_equal(y_sub_m.x, y_sub.x)


@requires_mne_sample_data
def test_xhemi():
    y = datasets.get_mne_stc(ndvar=True)
    data_dir = mne.datasets.sample.data_path()
    subjects_dir = os.path.join(data_dir, 'subjects')
    load.update_subjects_dir(y, subjects_dir)

    lh, rh = xhemi(y, mask=False)
    assert lh.source.rh_n == 0
    assert rh.source.rh_n == 0
    assert lh.max() == pytest.approx(10.80, abs=1e-2)
    assert rh.max() == pytest.approx(7.91, abs=1e-2)

    # Surface source space
    ds = datasets.get_mne_sample(src='ico', hpf=1, fsaverage=True)
    y = ds['src']
    yl, yr = xhemi(y)
    assert yl.source == yr.source
    # FSAverage
    y = ds['srcm']
    yl, yr = xhemi(y)
    assert yl.source == yr.source
    # Subset
    y = ds['srcm'].sub(source=[f'{label}-{hemi}' for label in ('superiortemporal', 'transversetemporal') for hemi in ('lh', 'rh')])
    yl, yr = xhemi(y)
    assert yl.source == yr.source

    # Volume source space
    ds = datasets.get_mne_sample(src='vol', ori='vector', hpf=1)
    y = ds[0, 'src']
    with pytest.raises(NotImplementedError):
        xhemi(y)
    # make symmetric
    coords = list(map(tuple, y.source.coordinates))
    index = [r == 0 or (-r, a, s) in coords for r, a, s in coords]
    y = y.sub(source=np.array(index))
    # test xhemi
    yl, yr = xhemi(y)
    assert yl.source == yr.source
    # test vector mirroring
    r, a, s = coords[10]
    assert r  # make sure it's not on midline
    for i_orig, coords in enumerate(y.source.coordinates):
        if tuple(coords) == (r, a, s):
            break
    for i_flipped, coords in enumerate(yr.source.coordinates):
        if tuple(coords) == (-r, a, s):
            break
    assert_array_equal(yr.x[i_flipped], y.x[i_orig] * [[-1], [1], [1]])


@requires_mne_sample_data  # source space distance computation times out
def test_source_space():
    "Test SourceSpace dimension"
    data_dir = mne.datasets.sample.data_path()
    subjects_dir = os.path.join(data_dir, 'subjects')
    annot_path = os.path.join(subjects_dir, '%s', 'label', '%s.%s.annot')

    for subject in ['fsaverage', 'sample']:
        mne_src = datasets._mne_source_space(subject, 'ico-4', subjects_dir)
        vertices = [mne_src[0]['vertno'], mne_src[1]['vertno']]
        ss = SourceSpace(vertices, subject, 'ico-4', subjects_dir)

        # labels
        for hemi_vertices, hemi in zip(ss.vertices, ('lh', 'rh')):
            labels, _, names = read_annot(annot_path % (subject, hemi, 'aparc'))
            start = 0 if hemi == 'lh' else len(ss.lh_vertices)
            hemi_tag = '-' + hemi
            for i, v in enumerate(hemi_vertices, start):
                label = labels[v]
                if label == -1:
                    assert ss.parc[i] == 'unknown' + hemi_tag
                else:
                    assert ss.parc[i] == names[label].decode() + hemi_tag

        # adjacency
        conn = ss.adjacency()
        mne_conn = mne.spatial_src_adjacency(mne_src)
        assert_array_equal(conn, _matrix_graph(mne_conn))

        # sub-space adjacency
        sssub = ss[ss._array_index('superiortemporal-rh')]
        ss2 = SourceSpace(vertices, subject, 'ico-4', subjects_dir, 'aparc')
        ss2sub = ss2[ss2._array_index('superiortemporal-rh')]
        assert_array_equal(sssub.adjacency(), ss2sub.adjacency())


@requires_mne_sample_data
def test_source_ndvar():
    "Test NDVar with source dimension"
    ds = datasets.get_mne_sample(-0.1, 0.1, src='ico', sub='index<=1')
    v = ds['src', 0]
    assert v.source.parc.name == 'aparc'
    v_2009 = set_parc(v, 'aparc.a2009s')
    assert v_2009.source.parc.name == 'aparc.a2009s'
    conn = v_2009.source.adjacency()
    assert np.sum(v.source.parc == v_2009.source.parc) < len(v.source)
    v_back = set_parc(v_2009, 'aparc')
    assert v_back.source.parc.name == 'aparc'
    assert_array_equal(v.source.parc, v_back.source.parc)
    assert v.x is v_back.x
    assert_array_equal(v_back.source.adjacency(), conn)

    # labels_from_cluster
    v1, v2 = ds['src']
    v1 = v1 * (v1 > 15)
    labels1 = labels_from_clusters(v1)
    assert len(labels1) == 1
    labels1s = labels_from_clusters(v1.sum('time'))
    assert len(labels1s) == 1
    assert_label_equal(labels1s[0], labels1[0])
    v2 = v2 * (v2 > 2)
    labels2 = labels_from_clusters(concatenate((v1, v2), 'case'))
    assert len(labels2) == 2
    assert_label_equal(labels1[0], labels2[0])


@requires_mne_sample_data
def test_vec_source():
    "Test vector source space"
    ds = datasets.get_mne_sample(0, 0.1, (0, 0), src='vol', sub="(modality=='A') & (side == 'L')", ori='vector', stc=True)
    # conversion: vector
    stc = ds[0, 'stc']
    stc2 = load.mne.stc_ndvar([stc, stc], ds.info['subject'], 'vol-10', ds.info['subjects_dir'])
    assert_dataobj_equal(stc2[1], ds[0, 'src'], name=False)
    # add parc
    stc2 = load.mne.stc_ndvar(stc, ds.info['subject'], 'vol-10', ds.info['subjects_dir'], parc='aparc+aseg')
    assert stc2.source.parc[0] == 'Right-Cerebellum-Cortex'
    # non-vector
    if hasattr(stc, 'magnitude'):  # added in mne 0.18
        stc = stc.magnitude()
        ndvar = load.mne.stc_ndvar(stc, ds.info['subject'], 'vol-10', ds.info['subjects_dir'])
        assert_dataobj_equal(ndvar, ds[0, 'src'].norm('space'), name=False)
    # test
    res = testnd.Vector('src', data=ds, samples=2)
    clusters = res.find_clusters()
    assert_array_equal(clusters['n_sources'], [808, 1, 8, 3, 1, 1])
    # NDVar
    v = ds['src']
    assert v.sub(source='lh', time=0).shape == (72, 712, 3)
    # parc
    v = ds[0, 'src']
    v = set_parc(v, Factor('abcdefg', repeat=227))
    v1 = v.sub(source='a')
    assert len(v1.source) == 227
    v2 = v.sub(source=('b', 'c'))
    assert len(v2.source) == 454
    assert 'b' in v2.source.parc
    assert 'd' not in v2.source.parc
    with pytest.raises(IndexError):
        v.sub(source='ab')
    with pytest.raises(IndexError):
        v.sub(source=['a', 'bc'])



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-generate-ndvar:

Creating NDVars
===============

.. currentmodule:: eelbrain

Shows how to initialize an :class:`NDVar` with the structure of EEG data from
(randomly generate) data arrays. The data is intended for illustrating EEG
analysis techniques and meant to vaguely resemble data from an N400 experiment,
but it is not meant to be a physiologically realistic simulation.

.. contents:: Contents
   :local:

"""
# sphinx_gallery_thumbnail_number = 3
import numpy as np
import scipy.spatial
from eelbrain import *


###############################################################################
# NDVars from arrays
# ------------------
# An :class:`NDVar` combines an n-dimensional :class:`numpy.ndarray` with
# :class:`Dimension` objects that describe what the
# different data axes mean, and provide meta information that is used, for
# example, for plotting.
# Here we start by create a Sensor dimension from a built-in EEG montage
# (a montage pairs sensor names with spatial locations on the head surface):
sensor = Sensor.from_montage('standard_alphabetic')
p = plot.SensorMap(sensor)

###############################################################################
# The dimension also contains information about the adjacency of its elements
# (i.e., specifying which elements are adjacent), which is used, for example,
# for cluster-based analysis. This information is imported automatically from
# :mod:`mne` when available; otherwise it can be defined manually when creating
# the sensor object, or based on pairwise sensor distance, as here:
sensor.set_adjacency(connect_dist=1.66)
p = plot.SensorMap(sensor, adjacency=True)

###############################################################################
# Using information from the :class:`Sensor` description about sensor coordinates, we
# can now generate an N400-like topography. After associating the data array
# with the Sensor description by creating an :class:`NDVar`, the topography can be plotted
# without any further information:
i_cz = sensor.names.index('Cz')
cz_loc = sensor.locations[i_cz]
dists = scipy.spatial.distance.cdist([cz_loc], sensor.locations)[0]
dists /= dists.max()
topo = -0.7 + dists
n400_topo = NDVar(topo, sensor)
p = plot.Topomap(n400_topo, clip='circle')

###############################################################################
# A time axis is specified using a :class:`UTS` ("uniform time series")
# object. As with the topography, the UTS object allows the :class:`NDVar` to
# automatically format the time axis of a figure. Here we create a simple time
# series based on a Gaussian:
window_data = scipy.signal.windows.gaussian(200, 12)[:140]
time = UTS(tmin=-0.100, tstep=0.005, nsamples=140)
n400_timecourse = NDVar(window_data, time)
p = plot.UTS(n400_timecourse)

###############################################################################
# Combining NDVars
# ---------------
# More complex NDVars can often be created by combining simpler NDVars.
# As example data, we generate random values for an independent variable
# (consistent with simulating an N400 response, we call it "cloze probability")
rng = np.random.RandomState(0)
n_trials = 100
cloze = np.concatenate([
    rng.uniform(0, 0.3, n_trials // 2),
    rng.uniform(0.8, 1.0, n_trials // 2),
])
rng.shuffle(cloze)
p = plot.Histogram(cloze)

###############################################################################
# Stacking NDVars
# ---------------
# A simple way of combining multiple NDVars is stacking them. Here we generate
# a separate topography for each cloze value, add some random noise, and then
# stack the resulting NDVars using :func:`combine`.
# The resulting stacked :class:`NDVar` has a :class:`Case` dimension reflecting
# the different cases (or trials):

ndvars = []
for cloze_i in cloze:
    topo_i = NDVar(topo * cloze_i + rng.normal(0, .5, len(topo)), sensor)
    ndvars.append(topo_i)
topographies = combine(ndvars)
topographies

###############################################################################
# The resulting NDvar can directly be used for a statistical test:

result = testnd.TTestOneSample(topographies)
p = plot.Topomap(result, clip='circle')
result

###############################################################################
# Casting NDVars
# --------------
# Multi-dimensional NDVars can also be created through multiplication of NDVars
# with different dimensions.
# Here, we put all the dimensions together to simulate the EEG signal. On the first
# line, turn cloze into :class:`Var` to make clear that cloze represents a
# :class:`Case` dimension, i.e. different trials:
signal = Var(1 - cloze) * n400_timecourse * n400_topo  

# Add noise
noise = powerlaw_noise(signal, 1)
noise = noise.smooth('sensor', 0.02, 'gaussian')
signal += noise

# Apply the average mastoids reference
signal -= signal.mean(sensor=['M1', 'M2'])

# Store EEG data in a Dataset with trial information
ds = Dataset({
    'eeg': signal,
    'cloze': Var(cloze),
    'predictability': Factor(cloze > 0.5, labels={True: 'high', False: 'low'}),
})

###############################################################################
# Plot the average simulated response
p = plot.TopoButterfly('eeg', data=ds, vmax=1.5, clip='circle', frame='t', axh=3)
p.set_time(0.400)

###############################################################################
# Plot averages separately for high and low cloze
p = plot.TopoButterfly('eeg', 'predictability', data=ds, vmax=1.5, clip='circle', frame='t', axh=3)
p.set_time(0.400)

###############################################################################
# Average over time in the N400 time window
p = plot.Topomap('eeg.mean(time=(0.300, 0.500))', 'predictability', data=ds, vmax=1, clip='circle')

###############################################################################
# Plot the first 20 trials, labeled with cloze propability
labels = [f'{i} ({c:.2f})' for i, c in enumerate(cloze[:20])]
p = plot.Butterfly('eeg[:20]', '.case', data=ds, axtitle=labels)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from eelbrain.testing import assert_dataobj_equal
from eelbrain import Categorial, Dataset, Factor, NDVar, Scalar, UTS, Var, datasets, table, combine


def test_cast_to_ndvar():
    "Test table.cast_to_ndvar()"
    long_ds = datasets.get_uv()
    long_ds['scalar'] = long_ds['A'] == 'a2'
    long_ds['time'] = long_ds.eval('A%B').as_var({
        ('a1', 'b1'): 0.,
        ('a1', 'b2'): 0.1,
        ('a2', 'b1'): 0.2,
        ('a2', 'b2'): 0.3,
    })

    # categorial
    ds = table.cast_to_ndvar('fltvar', 'A', 'B%rm', data=long_ds, name='new')
    assert ds.n_cases == long_ds.n_cases / 2
    assert ds['new'].A == Categorial('A', ('a1', 'a2'))

    # scalar
    ds2 = table.cast_to_ndvar('fltvar', 'scalar', 'B%rm', data=long_ds, dim='newdim', name='new')
    assert ds2.n_cases == long_ds.n_cases / 2
    assert ds2['new'].newdim == Scalar('newdim', [False, True])
    assert_array_equal(ds['new'].x, ds2['new'].x)

    # time
    ds = table.cast_to_ndvar('fltvar', 'time', 'rm', data=long_ds, dim='uts', name='y')
    assert ds.n_cases == long_ds.n_cases / 4
    assert ds['y'].time == UTS(0, 0.1, 4)


def test_difference():
    "Test table.difference"
    ds = datasets.get_uv()
    # add a variables that should stay in the dataset
    labels = {c: c[-1] for c in ds['rm'].cells}
    ds['rmf'] = Factor(ds['rm'], labels=labels)

    dds = table.difference('fltvar', 'A', 'a1', 'a2', 'rm', data=ds)
    assert repr(dds) == "<Dataset (20 cases) 'rm':F, 'fltvar':V, 'rmf':F>"
    assert_array_equal(dds['rmf'], Factor(dds['rm'], labels=labels))
    dds = table.difference('fltvar', 'A', 'a1', 'a2', 'rm % B', data=ds)
    assert repr(dds) == "<Dataset (40 cases) 'rm':F, 'B':F, 'fltvar':V, 'rmf':F>"
    # difference of the difference
    ddds = table.difference('fltvar', 'B', 'b1', 'b2', 'rm', data=dds)
    assert repr(ddds) == "<Dataset (20 cases) 'rm':F, 'fltvar':V, 'rmf':F>"
    dds = table.difference('fltvar', 'A%B', ('a1', 'b1'), ('a2', 'b2'), 'rm', data=ds)
    assert repr(dds) == "<Dataset (20 cases) 'rm':F, 'fltvar':V, 'rmf':F>"

    # sub
    dds = table.difference('fltvar', 'A', 'a1', 'a2', 'rm', sub="B == 'b1'", data=ds)
    assert repr(dds) == "<Dataset (20 cases) 'rm':F, 'fltvar':V, 'B':F, 'rmf':F>"

    # create bigger dataset
    ds2 = ds.copy()
    ds['C', :] = 'c1'
    ds2['C', :] = 'c2'
    ds = combine((ds, ds2))
    dds = table.difference('fltvar', 'A', 'a1', 'a2', 'rm % B % C', data=ds)
    assert repr(dds) == "<Dataset (80 cases) 'rm':F, 'B':F, 'C':F, 'fltvar':V, 'rmf':F>"
    dds = table.difference('fltvar', 'A%B', ('a1', 'b1'), ('a2', 'b2'), 'rm % C', data=ds)
    assert repr(dds) == "<Dataset (40 cases) 'rm':F, 'C':F, 'fltvar':V, 'rmf':F>"


def test_frequencies():
    "Test table.frequencies"
    ds = datasets.get_uts()
    freq = table.frequencies('YCat', 'A', data=ds)
    assert str(freq) == ('#   A    c1   c2   c3\n'
                         '---------------------\n'
                         '0   a0   10   10   10\n'
                         '1   a1   7    14   9 ')
    freq = table.frequencies('YCat', 'A % B', data=ds)
    assert str(freq) == ('#   A    B    c1   c2   c3\n'
                         '--------------------------\n'
                         '0   a0   b0   5    6    4 \n'
                         '1   a0   b1   5    4    6 \n'
                         '2   a1   b0   5    6    4 \n'
                         '3   a1   b1   2    8    5 ')
    freq = table.frequencies('YCat % A', 'B', data=ds)
    assert str(freq) == ('#   B    c1_a0   c1_a1   c2_a0   c2_a1   c3_a0   c3_a1\n'
                         '------------------------------------------------------\n'
                         '0   b0   5       5       6       6       4       4    \n'
                         '1   b1   5       2       4       8       6       5    ')
    ds = Dataset()
    ds['n'] = Factor('xxxyyy')
    freq = table.frequencies('n', data=ds)
    assert str(freq) == ('#   n   n_\n'
                         '----------\n'
                         '0   x   3 \n'
                         '1   y   3 ')


def test_melt_ndvar():
    "Test table.melt_ndvar()"
    ds = datasets.get_uts(True)
    ds = ds.sub("A == 'a1'")

    lds = table.melt_ndvar('uts', ds=ds)
    assert 'time' in lds
    assert isinstance(lds['time'], Var)
    assert_array_equal(np.unique(lds['time'].x), ds['uts'].time)

    # no ds
    lds2 = table.melt_ndvar(ds['uts'])
    assert_dataobj_equal(lds2['uts'], lds['uts'])
    # no ds no case
    lds3 = table.melt_ndvar(ds[0, 'utsnd'], 'sensor')
    assert_dataobj_equal(lds3[0, 'utsnd'], ds[0, 'utsnd'][0])

    # sensor
    lds = table.melt_ndvar("utsnd.summary(time=(0.1, 0.2))", ds=ds, varname='summary')
    assert set(lds['sensor'].cells) == set(ds['utsnd'].sensor.names)

    # NDVar out
    lds = table.melt_ndvar("utsnd", 'sensor', ds=ds)
    assert 'utsnd' in lds
    assert isinstance(lds['utsnd'], NDVar)
    assert_dataobj_equal(lds[:ds.n_cases, 'utsnd'], ds.eval("utsnd.sub(sensor='0')"))

    # more than one dimensions
    with pytest.raises(ValueError):
        table.melt_ndvar('utsnd', ds=ds)


def test_repmeas():
    "Test table.repmeas (repeated measures table)"
    ds = datasets.get_uv()
    print(table.repmeas('fltvar', 'A', 'rm', data=ds))
    print(table.repmeas('fltvar', 'A%B', 'rm', data=ds))
    print(table.repmeas('fltvar', 'A', 'B%rm', data=ds))

    # with int model
    ds['Bv'] = ds['B'].as_var({'b1': 1, 'b2': 2})
    print(table.repmeas('fltvar', 'A', 'Bv%rm', data=ds))

    # test naturalization of cellnames
    ds['ANum'] = Factor(ds['A'], labels={'a1': '1', 'a2': '2'})
    print(table.repmeas('fltvar', 'ANum', 'rm', data=ds))

    # with empty cell name
    ds['A'].update_labels({'a1': ''})
    print(table.repmeas('fltvar', 'A', 'rm', data=ds))



# Helper to download data for Alice example
from pathlib import Path

import pooch

from .._types import PathArg


def get_alice_path(
        path: PathArg = Path("~/Data/Alice"),
):
    path = Path(path).expanduser().resolve()
    path.mkdir(exist_ok=True, parents=True)

    baseurl = 'https://drum.lib.umd.edu/bitstream/handle/1903/27591/'
    registry = {
        'stimuli.zip': '92317dbfc81d6aef14fc334abd75d1165cf57501f0c11f8db1a47c76c3d90ac6',
        'eeg.1.zip': 'a645e4bf30ec8de10c92f82e9f842dd8172a4871f8eb23244e7e78b7dff157aa'
    }
    fetcher = pooch.Pooch(
        path=path,
        base_url=baseurl,
        registry=registry,
        retry_if_failed=4,
    )
    for fname in registry.keys():
        if (path / fname.split('.')[0]).exists():   # Won't work for multiple eeg.x.zip download
            continue
        fetcher.fetch(fname, processor=pooch.Unzip(extract_dir='.'))
        (path / fname).unlink()
    return path



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import pytest

pytest.register_assert_rewrite('eelbrain.testing._testing')

from ..datasets._simple import get_ndvar
from ._testing import (
    gui_test, hide_plots,
    requires_framework_build, requires_mne_sample_data, requires_mne_testing_data, requires_r_ez,
    skip_on_windows,
    TempDir, working_directory,
    assert_dataset_equal, assert_dataobj_equal, assert_fmtxt_str_equals, assert_source_space_equal,
    file_path, path,
)



"""Interface for TeX command line tools

Adapted from the obsolete ``tex`` module by Volker Grabsch


Original Notes
--------------
Temporary files are always cleaned up.
The TeX interpreter is automatically re-run as often as necessary,
and an exception is thrown
in case the output fails to stabilize soon enough.
The TeX interpreter is always run in batch mode,
so it won't ever get in your way by stopping your application
when there are issues with your TeX source.
Instead, an exception is thrown
that contains all information of the TeX log.


License
-------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject
to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import subprocess
import tempfile


def convert(tex_source, input_format, output_format, max_runs=5):
    '''Convert LaTeX or TeX source to PDF or DVI.'''
    # check arguments
    assert isinstance(tex_source, str)
    try:
        (tex_cmd, output_suffix) = {
            ('tex',   'dvi'): ('tex',      '.dvi'),
            ('latex', 'dvi'): ('latex',    '.dvi'),
            ('tex',   'pdf'): ('pdftex',   '.pdf'),
            ('latex', 'pdf'): ('pdflatex', '.pdf'),
        }[(input_format, output_format)]
    except KeyError:
        raise ValueError('Unable to handle conversion: %s -> %s'
                         % (input_format, output_format))
    if max_runs < 2:
        raise ValueError('max_runs must be at least 2.')
    # create temporary directory
    with tempfile.TemporaryDirectory(prefix='tex-temp-') as tex_dir:
        # create LaTeX source file
        tex_filename = os.path.join(tex_dir, 'texput.tex')
        with open(tex_filename, 'w') as fid:
            fid.write(tex_source)

        # run LaTeX processor as often as necessary
        aux_old = None
        for i in range(max_runs):
            tex_process = subprocess.Popen(
                [tex_cmd,
                    '-interaction=batchmode',
                    '-halt-on-error',
                    '-no-shell-escape',
                    tex_filename,
                 ],
                stdin=open(os.devnull, 'r'),
                stdout=open(os.devnull, 'w'),
                stderr=subprocess.STDOUT,
                close_fds=True,
                shell=False,
                cwd=tex_dir,
                env={'PATH': os.getenv('PATH')},
            )
            tex_process.wait()
            if tex_process.returncode != 0:
                with open(os.path.join(tex_dir, 'texput.log'), 'rb') as fid:
                    log = fid.read()
                raise ValueError(log.decode())

            with open(os.path.join(tex_dir, 'texput.aux'), 'rb') as fid:
                aux = fid.read()

            if aux == aux_old:  # aux file stabilized
                output_filename = os.path.join(tex_dir, 'texput' + output_suffix)
                if not os.path.exists(output_filename):
                    raise RuntimeError('No output file was produced.')
                else:
                    with open(output_filename, 'rb') as fid:
                        return fid.read()
            aux_old = aux
            # TODO:
            # Also handle makeindex and bibtex,
            # possibly in a similar manner as described in:
            # http://vim-latex.sourceforge.net/documentation/latex-suite/compiling-multiple.html
        raise RuntimeError("%s didn't stabilize after %i runs" % ('texput.aux', max_runs))


def tex2dvi(tex_source, **kwargs):
    '''Convert TeX source to DVI.'''
    return convert(tex_source, 'tex', 'dvi', **kwargs)


def latex2dvi(tex_source, **kwargs):
    '''Convert LaTeX source to DVI.'''
    return convert(tex_source, 'latex', 'dvi', **kwargs)


def tex2pdf(tex_source, **kwargs):
    '''Convert TeX source to PDF.'''
    return convert(tex_source, 'tex', 'pdf', **kwargs)


def latex2pdf(tex_source, **kwargs):
    '''Convert LaTeX source to PDF.'''
    return convert(tex_source, 'latex', 'pdf', **kwargs)



# skip test: data unavailable
from eelbrain.pipeline import *


class Mouse(MneExperiment):

    # Name of the experimental session(s), used to locate *-raw.fif files
    sessions = 'CAT'

    # Pre-processing pipeline: each entry in `raw` specifies one processing step. The first parameter
    # of each entry specifies the source (another processing step or 'raw' for raw input data).
    raw = {
        # Maxwell filter as first step (taking input from raw data, 'raw')
        'tsss': RawMaxwell('raw', st_duration=10., ignore_ref=True, st_correlation=0.9, st_only=True),
        # Band-pass filter data between 1 and 40 Hz (taking Maxwell-filtered data as input, 'tsss)
        '1-40': RawFilter('tsss', 1, 40),
        # Perform ICA on filtered data
        'ica': RawICA('1-40', 'CAT', n_components=0.99),
    }

    # Variables determine how event triggeres are mapped to meaningful labels. Events are represented
    # as data-table in which each row corresponds to one event (i.e., one trigger). Each variable
    # defined here adds one column in that data-table, assigning a label or value to each event.
    variables = {
        # The first parameter specifies the source variable (here the trigger values),
        # the second parameter a mapping from source to target labels/values
        'stimulus': LabelVar('trigger', {(162, 163): 'target', (166, 167): 'prime'}),
        'prediction': LabelVar('trigger', {(162, 166): 'expected', (163, 167): 'unexpected'}),
    }

    # Epochs specify how to extract time-locked data segments ("epochs") from the continuous data.
    epochs = {
        # A PrimaryEpoch definition extracts epochs directly from continuous data. The first argument
        # specifies the recording session from which to extract the data (here: 'CAT'). The second
        # argument specifies which events to extract the data from (here: all events at which the
        # 'stimulus' variable, defined above, has a value of either 'prime' or 'target').
        'word': PrimaryEpoch('CAT', "stimulus.isin(('prime', 'target'))", samplingrate=200),
        # A secondary epoch inherits its properties from the base epoch ("word") unless they are
        # explicitly modified (here, selecting a subset of events)
        'prime': SecondaryEpoch('word', "stimulus == 'prime'"),
        'target': SecondaryEpoch('word', "stimulus == 'target'"),
        # The 'cov' epoch defines the data segments used to compute the noise covariance matrix for
        # source localization
        'cov': SecondaryEpoch('prime', tmax=0),
    }

    # Tests define contrasts or comparisons based on variables.
    # In the example, the following would test for the difference between expected and unexpected words in
    # target words:
    # >>> result = mouse.load_test(test='surprise', epoch='target')
    tests = {
        '=0': TTestOneSample(),
        'surprise': TTestRelated('prediction', 'unexpected', 'expected'),
        'anova': ANOVA('prediction * subject'),
    }

    # Define masks for permutation tests in anatomically constrained regions. Fro exmample:
    # >>> result = mouse.load_test(test='surprise', epoch='target', mask='lateraltemporal')
    parcs = {
        'STG': SubParc('aparc', ('transversetemporal', 'superiortemporal')),
        'IFG': SubParc('aparc', ('parsopercularis', 'parsorbitalis', 'parstriangularis')),
        'lateraltemporal': SubParc('aparc', (
            'transversetemporal', 'superiortemporal', 'bankssts',
            'middletemporal', 'inferiortemporal')),
    }


# Initialize the pipeline instance with the root path where the data is located
root = '~/Data/Mouse'
e = Mouse(root)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-two-stage:

Two-stage test
==============

.. currentmodule:: eelbrain

When trials are associated with continuous predictor variables, averaging is
often a poor solution that loses part of the data. In such cases, a two-stage
design can be employed that allows using the continuous predictor variable to
test hypotheses at the group level. A two-stage analysis involves:

 - Stage 1: fit a regression model to each individual subject's data
 - Stage 2: test regression coefficients at the group level

The example uses the same simulated data and design used in :ref:`exa-lm`. 
The data are meant to vaguely resemble data from a word reading
experiment, but not intended as a physiologically realistic simulation.
"""
# sphinx_gallery_thumbnail_number = 1
from eelbrain import *

###############################################################################
# Stage 1
# -------
# Generate simulated data: each function call to :func:`datasets.simulate_erp`
# generates a dataset for one subject (in a real experiment this would be 
# replaced with a function that loads data for this subject). 
# For each subject, a multiple regression model is fit using n characters and 
# cloze probability as continuous predictor variables.
lms = []
for subject in range(10):
    # generate data for one subject
    ds = datasets.simulate_erp(seed=subject)
    # Re-reference EEG data
    ds['eeg'] -= ds['eeg'].mean(sensor=['M1', 'M2'])
    # Fit stage 1 model (samples=0 because we do not need permutations at stage 1)
    lm = testnd.LM('eeg', 'n_chars + cloze', data=ds, samples=0, subject=str(subject))
    lms.append(lm)

###############################################################################
# Stage 2
# -------
# Prepare a :class:`Dataset` with the first level statistic of interest. 

rows = []
for lm in lms:
    rows.append([lm.subject, lm.t('intercept'), lm.t('n_chars'), lm.t('cloze')])
# When creating the dataset for stage 2 analysis, declare subject as random factor;
# this is only relevant if performing ANOVA as stage 2 test.
data = Dataset.from_caselist(['subject', 'intercept', 'n_chars', 'cloze'], rows, random='subject')
data

###############################################################################
# Now we can test whether the first stage estimates are consistent across subject.

result = testnd.TTestOneSample('n_chars', data=data, pmin=0.05, tstart=0, tstop=0.300)
p = plot.TopoArray(result, t=[0.120, 0.155, None], title=result, head_radius=0.35)
p_cb = p.plot_colorbar(right_of=p.axes[0], label='t')

###############################################################################
# Instead of *t*-values, we might want to visualize regression coefficients:

rows = []
for lm in lms:
    rows.append([lm.subject, lm.coefficient('n_chars')])
data_c = Dataset.from_caselist(['subject', 'n_chars'], rows, random='subject')

# mask regression coefficients by significance to add outlines to plot
masked_c = data_c['n_chars'].mean('case').mask(result.p > 0.05, missing=True)

p = plot.TopoArray(masked_c, t=[0.120, 0.155, None], title=result, head_radius=0.35)
p_cb = p.plot_colorbar(right_of=p.axes[0], label='µV', unit=1e-6)

###############################################################################
# Of course, other tests could be applied at stage 2, for example 
#  * *T*-tests to compare coefficients for two different regressor, or two
#    differen subject groups
#  * ANOVA for multiple regressors and/or subject groups
#  * Multiple regression models with subject variables to test for individual
#    differnces



"""Some basic statistical tests for univariate data."""
__test__ = False
from ._stats.glm import ANOVA
from ._stats.test import (
    Correlation, RankCorrelation,
    TTestOneSample, TTestIndependent, TTestRelated, MannWhitneyU, WilcoxonSignedRank,
    pairwise, ttest, correlations,
    lilliefors,
    pairwise_correlations,
)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Simulate EEG data"""
import numpy as np
import scipy.signal
import scipy.spatial

from .._data_obj import Dataset, Factor, Var, NDVar, Sensor, UTS
from .._ndvar import gaussian, powerlaw_noise
from .. import _info


def _topo(sensor, center):
    i = sensor.names.index(center)
    loc = sensor.locations[i]
    dists = scipy.spatial.distance.cdist([loc], sensor.locations)[0]
    radius = sensor._sphere_fit[1].mean()
    dists /= radius
    topo = 1.0 - dists
    return NDVar(topo, (sensor,))


def simulate_erp(
        n_trials: int = 80,
        seed: int = 0,
        snr: float = 0.2,
        sensor: Sensor = None,
        time: UTS = None,
        short_long: bool = False,
) -> Dataset:
    """Simulate event-related EEG data

    Parameters
    ----------
    n_trials
        Number of trials (default 100).
    seed
        Random seed.
    snr
        Signal-to-noise ratio.
    sensor
        Sensor dimension.
    time
        Time dimension.
    short_long
        Create bimodal distribution of word length for categorial analysis.

    Examples
    --------
    Compare with kiloword::

        ys = datasets.simulate_erp()['eeg']
        ys -= ys.mean(sensor=['M1', 'M2'])
        import mne
        path = mne.datasets.kiloword.data_path()
        y = load.mne.epochs_ndvar(path + '/kword_metadata-epo.fif')
        plot.TopoButterfly([y, ys])

    """
    assert n_trials % 2 == 0

    if sensor is None:
        sensor = Sensor.from_montage('standard_alphabetic')
        sensor.set_adjacency(connect_dist=1.66)
    if time is None:
        time = UTS(-0.100, 0.005, 140)

    # Generate random values for the independent variable (cloze probability)
    rng = np.random.RandomState(seed)
    cloze_x = np.concatenate([
        rng.uniform(0, 0.3, n_trials // 2),
        rng.uniform(0.8, 1.0, n_trials // 2),
    ])
    # Word length (number of characters)
    if short_long:
        assert not n_trials % 4
        n_condition = n_trials // 4
        params = [(4, 1.5), (8, 2), (4, 1.5), (8, 2)]
        n_chars = np.concatenate([rng.normal(loc, scale, n_condition) for loc, scale in params])
        n_chars = np.clip(n_chars, 2, None)
    else:
        rng.shuffle(cloze_x)
        n_chars = rng.uniform(3, 7, n_trials)
    cloze = Var(cloze_x)
    n_chars = Var(np.round(n_chars)).astype(int)

    # Generate topography
    n400_topo = -2.0 * _topo(sensor, 'Cz')
    # Generate timing
    n400_timecourse = gaussian(0.400, 0.034, time)
    # Put all the dimensions together to simulate the EEG signal
    surprisal = -cloze.log(2)
    signal = surprisal * 0.25 * n400_timecourse * n400_topo

    # add early responses:
    # 130 ms
    tc = gaussian(0.130, 0.025, time)
    topo = _topo(sensor, 'O1') + _topo(sensor, 'O2') - 0.5 * _topo(sensor, 'Cz')
    signal += (n_chars * 0.5) * tc * topo
    # 195 ms
    amp = Var(rng.normal(0.8, 1, n_trials))
    tc = gaussian(0.195, 0.015, time)
    topo = 1.2 * _topo(sensor, 'F3') + _topo(sensor, 'F4')
    signal += amp * tc * topo
    # 270
    amp = Var(rng.normal(1., 1, n_trials))
    tc = gaussian(0.270, 0.050, time)
    topo = _topo(sensor, 'O1') + _topo(sensor, 'O2')
    signal += amp * tc * topo
    # 280
    amp = Var(rng.normal(-1, 1, n_trials))
    tc = gaussian(0.280, 0.030, time)
    topo = _topo(sensor, 'Pz')
    signal += amp * tc * topo
    # 600
    amp = Var(rng.normal(0.5, 1, n_trials))
    tc = gaussian(0.590, 0.100, time)
    topo = -_topo(sensor, 'Fz')
    signal += amp * tc * topo

    # Add noise
    noise = powerlaw_noise(signal, 1, rng)
    noise = noise.smooth('sensor', 0.02, 'gaussian')
    noise *= (signal.std() / noise.std() / snr)
    signal += noise

    # Data scale
    signal *= 1e-6
    signal.info = _info.for_eeg()

    # Store EEG data in a Dataset with trial information
    ds = Dataset()
    ds['eeg'] = signal
    ds['cloze'] = Var(cloze_x)
    ds['predictability'] = Factor(cloze_x > 0.5, labels={True: 'high', False: 'low'})
    ds['n_chars'] = Var(n_chars)
    if short_long:
        ds['length'] = Factor(['short', 'long'], repeat=n_condition, tile=2)
    return ds



# skip test: data unavailable
from pathlib import Path

import mne.coreg
# from mne.coreg import read_mri_cfg

root = Path('~/Data/Mouse').expanduser()
mri_dir = root / 'mri'
# mne.coreg.create_default_subject('/Applications', subjects_dir=mri_dir)
for subject_path in mri_dir.glob('S*'):
    cfg = mne.coreg.read_mri_cfg(subject_path.name, mri_dir)
    scale = cfg['scale']
    mne.scale_mri('fsaverage', subject_path.name, scale, overwrite=True, subjects_dir=mri_dir, labels=False)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"Utilities for testing"
from contextlib import ContextDecorator, contextmanager
from functools import reduce
import os
from operator import mul
from pathlib import Path
import shutil
import tempfile

import mne
import numpy as np
from numpy.testing import assert_array_equal
import pytest

import eelbrain._wxgui
from .._config import CONFIG
from .._data_obj import Dataset, NDVar, Var, Factor, isdatalist, isuv


class TempDir(str):
    "After MNE-Python mne.utils"
    def __new__(cls):
        return str.__new__(cls, tempfile.mkdtemp())

    def __del__(self):
        shutil.rmtree(self, ignore_errors=True)


def assert_dataset_equal(ds1, ds2, decimal=None):
    """
    Raise an assertion if two Datasets are not equal up to desired precision.

    Parameters
    ----------
    ds1, ds2 : Dataset
        Datasets to compare.
    decimal : None | int
        Desired precision (default is exact match).
    """
    assert ds1.keys() == ds2.keys()
    for k in ds1.keys():
        assert_dataobj_equal(ds1[k], ds2[k], decimal=decimal)
    assert ds1.info.keys() == ds2.info.keys()


def assert_dataobj_equal(d1, d2, decimal=None, name=True):
    """Assert that two data-objects are equal up to desired precision.

    Parameters
    ----------
    d1, d2 : data-objects
        Data-objects to compare.
    decimal : None | int
        Desired precision (default is exact match).
    name : bool
        Assert that ``d1.name == d2.name``.
    """
    assert type(d1) == type(d2)  # noqa: E721
    if name:
        assert d1.name == d2.name
    assert len(d1) == len(d2)
    if isuv(d1, interaction=True):
        if isinstance(d1, Var):
            is_equal = np.isclose(d1.x, d2.x, equal_nan=True, rtol=0, atol=10**-decimal if decimal else 0)
        else:
            is_equal = d1 == d2
        if not np.all(is_equal):
            ds = Dataset()
            ds['value'] = d1.copy()
            ds['target'] = d2.copy()
            ds['unequal'] = Factor(is_equal, labels={True: '', False: 'x'})
            if isinstance(d1, Var):
                ds['difference'] = d1 - d2
            raise AssertionError(f'Two {d1.__class__.__name__}s named {d1.name!r} have unequal values:\n\n{ds}')
        if isinstance(d1, Factor):
            if d1.random != d2.random:
                raise AssertionError(f"{d1.name}: d1.random={d1.random}, d2.random={d2.random}")
    elif isinstance(d1, NDVar):
        assert d1.dims == d2.dims
        if decimal:
            is_different = np.max(np.abs(d1.x - d2.x)) >= 10**-decimal
        else:
            is_different = np.any(d1.x != d2.x)

        if is_different:
            n = reduce(mul, d1.x.shape)
            difference = np.abs(d1.x - d2.x)
            if decimal:
                different = difference >= 10**-decimal
            else:
                different = d1.x != d2.x
            n_different = different.sum()
            mean_diff = difference[different].sum() / n_different
            raise AssertionError(f"NDVars named {d1.name!r} have unequal values. Difference in {n_different} of {n} values, average difference={mean_diff}.")
    elif isdatalist(d1):
        assert all(item1 == item2 for item1, item2 in zip(d1, d2))
    elif isinstance(d1, Dataset):
        assert_dataset_equal(d1, d2, decimal)
    else:
        raise TypeError(f"{d1.__class__.__name__} is not a data-object type: {d1!r}")


def assert_source_space_equal(src1, src2):
    """Assert that two SourceSpace objects are identical

    Parameters
    ----------
    src1, src2 : SourceSpace objects
        SourceSpace objects to compare.
    """
    assert_array_equal(src1.vertices[0], src2.vertices[0])
    assert_array_equal(src1.vertices[1], src2.vertices[1])
    assert src1.subject == src2.subject
    assert src1.src == src2.src
    assert src1.subjects_dir == src2.subjects_dir


class GUITestContext(ContextDecorator):
    modules = (
        eelbrain._wxgui.select_epochs,
        eelbrain._wxgui.select_components,
        eelbrain._wxgui.history,
        eelbrain._wxgui.load_stcs,
    )

    def __init__(self):
        self._i = 0

    def __enter__(self):
        self._i += 1
        if self._i == 1:
            for mod in self.modules:
                mod.TEST_MODE = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._i -= 1
        if self._i == 0:
            for mod in self.modules:
                mod.TEST_MODE = False


gui_test_context = GUITestContext()


def gui_test(function):
    return gui_test_context(requires_framework_build(function))


def assert_fmtxt_str_equals(
        fmtxt,
        target: str,
        indent: int = 4,
):
    """Allow clean and PEP8 compliant target strings in testing"""
    as_str = str(fmtxt)
    as_str = '\n'.join(line.rstrip() for line in as_str.split('\n'))  # rstrip whitespace (for pep8)
    target = '\n'.join(line[indent:] for line in target.splitlines()[1:-1])
    assert as_str == target


class ConfigContext(ContextDecorator):

    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.old = None

    def __enter__(self):
        self.old = CONFIG[self.key]
        CONFIG[self.key] = self.value

    def __exit__(self, exc_type, exc_val, exc_tb):
        CONFIG[self.key] = self.old


hide_plots_context = ConfigContext('show', False)


def hide_plots(function):
    return hide_plots_context(requires_framework_build(function))


@contextmanager
def working_directory(wd):
    "Context for temporarily changing the working directory"
    cwd = os.getcwd()
    os.chdir(wd)
    try:
        yield
    finally:
        os.chdir(cwd)


def requires_framework_build(function):
    return pytest.mark.framework_build(function)


def requires_mne_sample_data(function):
    if mne.datasets.sample.data_path(download=False):
        return function
    else:
        return pytest.mark.skip('mne sample data unavailable')(function)


def requires_mne_testing_data(function):
    if mne.datasets.testing.data_path(download=False):
        return function
    else:
        return pytest.mark.skip('mne testing data unavailable')(function)


def requires_r_ez(function):
    from .._utils.r_bridge import r, r_warning_filter

    with r_warning_filter:
        success = r('require(ez)')[0]

    if success:
        return function
    else:
        return pytest.mark.skip('r-ez unavailable')(function)


def skip_on_windows(function):
    if os.name == 'nt':
        return pytest.mark.skip('Test disabled on Windows')(function)
    else:
        return function


def file_path(name):
    "Path to test data file in the test_data directory"
    root = Path(__file__).parents[2]
    if name == 'fox-prestige':
        path = root / 'examples' / 'statistics' / 'Fox_Prestige_data.txt'
    else:
        path = root / 'test_data' / name

    if path.exists():
        return path
    else:
        raise IOError("Testing file does not exist. Test can only be executed from source repository.")


def path(string):
    "OS-independent path check"
    return str(Path(string))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import pytest
from scipy import signal

from eelbrain import NDVar, Case, Scalar, UTS, datasets, concatenate, convolve, correlation_coefficient, cross_correlation, cwt_morlet, find_intervals, find_peaks, frequency_response, gaussian, normalize_in_cells, psd_welch, resample, set_time
from eelbrain.testing import assert_dataobj_equal, get_ndvar


def test_concatenate():
    """Test concatenate()

    Concatenation of SourceSpace is tested in .test_mne.test_source_estimate
    """
    ds = datasets.get_uts(True)

    v0 = ds[0, 'utsnd']
    v1 = ds[1, 'utsnd']
    vc = concatenate((v1, v0))
    assert_array_equal(vc.sub(time=(0, 1)).x, v1.x)
    assert_array_equal(vc.sub(time=(1, 2)).x, v0.x)
    assert_array_equal(vc.info, ds['utsnd'].info)

    # scalar
    psd = psd_welch(ds['utsnd'], n_fft=100)
    v0 = psd.sub(frequency=(None, 5))
    v1 = psd.sub(frequency=(45, None))
    conc = concatenate((v0, v1), 'frequency')
    assert_array_equal(conc.frequency.values[:5], psd.frequency.values[:5])
    assert_array_equal(conc.frequency.values[5:], psd.frequency.values[45:])
    conc_data = conc.get_data(v1.dimnames)
    assert_array_equal(conc_data[:, :, 5:], v1.x)

    # cat
    x = get_ndvar(2, frequency=0, cat=4)
    x_re = concatenate([x.sub(cat=(None, 'c')), x.sub(cat=('c', None))], 'cat')
    assert_dataobj_equal(x_re, x)


def test_convolve():
    # convolve is also tested in test_boosting.py
    ds = datasets._get_continuous()

    h1d = ds['h1']
    h2d = ds['h2']
    x1d = ds['x1']
    x2d = ds['x2']

    # 1d
    xc = convolve(h1d, x1d)
    xc_np = np.convolve(h1d.x, x1d.x)
    assert_array_equal(xc.x, xc_np[:100])

    # 2d
    xc = convolve(h2d, x2d)
    xc_np = np.convolve(h2d.x[0], x2d.x[0]) + np.convolve(h2d.x[1], x2d.x[1])
    assert_allclose(xc.x, xc_np[:100])

    # add dimension through kernel
    xc = convolve(h2d, x1d)
    xc_np = np.vstack((np.convolve(h2d.x[0], x1d.x), np.convolve(h2d.x[1], x1d.x)))
    assert_array_equal(xc.x, xc_np[:, :100])

    # add dimension through x
    xc = convolve(h1d, x2d)
    xc_np = np.vstack((np.convolve(h1d.x, x2d.x[0]), np.convolve(h1d.x, x2d.x[1])))
    assert_array_equal(xc.x, xc_np[:, :100])

    # 2 predictors
    xc = convolve([h1d, h2d], [x1d, x2d])
    xc_np = np.convolve(h1d.x, x1d.x) + np.convolve(h2d.x[0], x2d.x[0]) + np.convolve(h2d.x[1], x2d.x[1])
    assert_allclose(xc.x, xc_np[:100])


def test_correlation_coefficient():
    ds = datasets.get_uts()
    uts = ds['uts']
    uts2 = uts.copy()
    uts2.x += np.random.normal(0, 1, uts2.shape)

    assert correlation_coefficient(uts, uts2) == pytest.approx(
        np.corrcoef(uts.x.ravel(), uts2.x.ravel())[0, 1])
    assert_allclose(
        correlation_coefficient(uts[:10], uts2[:10], 'time').x,
        [np.corrcoef(uts.x[i], uts2.x[i])[0, 1] for i in range(10)])
    assert_allclose(
        correlation_coefficient(uts[:, :-.1], uts2[:, :-.1], 'case').x,
        [np.corrcoef(uts.x[:, i], uts2.x[:, i])[0, 1] for i in range(10)])


def test_cross_correlation():
    ds = datasets._get_continuous()
    x = ds['x1']

    assert cross_correlation(x, x).argmax() == 0
    assert cross_correlation(x[2:], x).argmax() == 0
    assert cross_correlation(x[:9], x).argmax() == 0
    assert cross_correlation(x, x[1:]).argmax() == 0
    assert cross_correlation(x, x[:8]).argmax() == 0
    assert cross_correlation(x[2:], x[:8]).argmax() == 0


def test_cwt():
    ds = datasets._get_continuous()
    # 1d
    y = cwt_morlet(ds['x1'], [2, 3, 4])
    assert y.ndim == 2
    # 2d
    y = cwt_morlet(ds['x2'], [2, 3, 4])
    assert y.ndim == 3


def test_diff():
    ds = datasets.get_uts()
    y = ds['uts']
    time_mask = np.logical_and(y.time.times > 0, y.time.times < 0.100).reshape((1, -1))
    mask = np.repeat(time_mask, len(y), 0)
    y_masked = y.mask(mask)
    # target for time-diff
    time_mask = np.logical_and(y.time.times > 0, y.time.times < 0.110).reshape((1, -1))
    target_mask = np.repeat(time_mask, len(y), 0)

    # 1d
    diff = y_masked[0].diff('time')
    assert_array_equal(diff.x.data, np.diff(y.x[0], 1, 0, y.x[0, 0]))
    assert_array_equal(diff.x.mask, target_mask[0])

    # 2d
    diff = y_masked.diff('case')
    assert_array_equal(diff.x.data, np.diff(y.x, 1, 0, y.x[:1]))
    assert_array_equal(diff.x.mask, mask)

    diff = y_masked.diff('time')
    assert_array_equal(diff.x.data, np.diff(y.x, 1, 1, y.x[:, :1]))
    assert_array_equal(diff.x.mask, target_mask)


def test_dot():
    ds = datasets.get_uts(True)

    # x subset of y
    index = ['3', '2']
    utsnd = ds['utsnd']
    topo = utsnd.mean(('case', 'time'))
    y1 = topo.sub(sensor=index).dot(utsnd.sub(sensor=index))
    assert_dataobj_equal(topo[index].dot(utsnd), y1)
    assert_dataobj_equal(topo.dot(utsnd.sub(sensor=index)), y1)


def test_find_intervals():
    time = UTS(-5, 1, 10)
    x = NDVar([0, 1, 0, 1, 1, 0, 1, 1, 1, 0], (time,))
    assert find_intervals(x) == ((-4, -3), (-2, 0), (1, 4))
    x = NDVar([0, 1, 0, 1, 1, 0, 1, 1, 1, 1], (time,))
    assert find_intervals(x) == ((-4, -3), (-2, 0), (1, 5))
    x = NDVar([1, 1, 0, 1, 1, 0, 1, 1, 1, 1], (time,))
    assert find_intervals(x) == ((-5, -3), (-2, 0), (1, 5))


def test_find_peaks():
    scalar = Scalar('scalar', range(9))
    time = UTS(0, .1, 12)
    v = NDVar.zeros((scalar, time))
    wsize = [0, 0, 1, 2, 3, 2, 1, 0, 0]
    for i, s in enumerate(wsize):
        if s:
            v.x[i, 5 - s: 5 + s] += np.hamming(2 * s)

    v = round(v, 5)  # numpy hamming window is not exactly symmetric
    peaks = find_peaks(v)
    assert_array_equal(np.where(peaks.x), ([4, 4], [4, 5]))


def test_frequency_response():
    b_array = signal.firwin(80, 0.5, window=('kaiser', 8))
    freqs_array, fresp_array = signal.freqz(b_array)
    hz_to_rad = 2 * np.pi * 0.01

    b = NDVar(b_array, (UTS(0, 0.01, 80),))
    fresp = frequency_response(b)
    assert_array_equal(fresp.x, fresp_array)
    assert_array_equal(fresp.frequency.values * hz_to_rad, freqs_array)

    b2d = concatenate((b, b), Case)
    fresp = frequency_response(b2d)
    assert_array_equal(fresp.x[0], fresp_array)
    assert_array_equal(fresp.x[1], fresp_array)
    assert_array_equal(fresp.frequency.values * hz_to_rad, freqs_array)


def test_gaussian():
    time = UTS(0, 0.1, 5)
    assert_array_equal(gaussian(0.1, 0.2, time).x, signal.windows.gaussian(7, 2)[2:])
    assert_array_equal(gaussian(0.2, 0.2, time).x, signal.windows.gaussian(5, 2))
    assert_array_equal(gaussian(0.3, 0.1, time).x, signal.windows.gaussian(7, 1)[:5])
    time = UTS(0, 0.1, 6)
    assert_array_equal(gaussian(0.1, 0.2, time).x, signal.windows.gaussian(9, 2)[3:])
    assert_array_equal(gaussian(0.2, 0.2, time).x, signal.windows.gaussian(9, 2)[2:8])
    assert_array_equal(gaussian(0.3, 0.2, time).x, signal.windows.gaussian(9, 2)[1:7])
    assert_array_equal(gaussian(0.4, 0.1, time).x, signal.windows.gaussian(9, 1)[:6])


def test_mask():
    ds = datasets.get_uts(True)

    x = NDVar([1, 2, 3], Case)
    assert x.mean() == 2.0
    y = x.mask([True, False, False])
    assert y.mean() == 2.5

    # multi-dimensional
    y = ds[:2, 'utsnd'].copy()
    mask_x = y.time.times >= 0.500
    mask_ndvar = NDVar(mask_x, y.time)
    y_masked = y.mask(mask_ndvar)
    assert_array_equal(y_masked.x.mask[:, :, 70:], True)
    assert_array_equal(y_masked.x.mask[:, :, :70], False)
    # mask that is smaller than array
    mask = mask_ndvar.sub(time=(0.100, None))
    with pytest.raises(ValueError):
        y.mask(mask)
    y_masked = y.mask(mask, missing=True)
    assert_array_equal(y_masked.x.mask[:, :, 70:], True)
    assert_array_equal(y_masked.x.mask[:, :, 30:70], False)
    assert_array_equal(y_masked.x.mask[:, :, :30], True)


def test_normalize_in_cells():
    ds = datasets.get_uts(True)
    ab = ds.eval("A % B")
    indices = [ab == cell for cell in ab.cells]
    # z-score
    ds['utsnd_n'] = normalize_in_cells('utsnd', 'sensor', data=ds)
    y_mean = ds['utsnd_n'].mean('case')
    assert_allclose(y_mean.mean('sensor'), 0, atol=1e-10)
    assert_allclose(y_mean.std('sensor'), 1)
    ds['utsnd_n'] = normalize_in_cells('utsnd', 'sensor', 'A % B', data=ds)
    for index in indices:
        y_mean = ds[index, 'utsnd_n'].mean('case')
        assert_allclose(y_mean.mean('sensor'), 0, atol=1e-10)
        assert_allclose(y_mean.std('sensor'), 1)
    # range
    ds['utsnd_n'] = normalize_in_cells('utsnd', 'sensor', data=ds, method='range')
    y_mean = ds['utsnd_n'].mean('case')
    assert_allclose(y_mean.min('sensor'), 0, atol=1e-10)
    assert_allclose(y_mean.max('sensor'), 1)
    ds['utsnd_n'] = normalize_in_cells('utsnd', 'sensor', 'A % B', data=ds, method='range')
    for index in indices:
        y_mean = ds[index, 'utsnd_n'].mean('case')
        assert_allclose(y_mean.min('sensor'), 0, atol=1e-10)
        assert_allclose(y_mean.max('sensor'), 1)


def test_resample():
    x = NDVar([0.0, 1.0, 1.4, 1.0, 0.0], UTS(0, 0.1, 5)).mask([True, False, False, False, True])
    y = resample(x, 20)
    assert_array_equal(y.x.mask, [True, False, False, False, False, False, False, False, True, True])
    y = resample(x, 20, npad=0)
    assert_array_equal(y.x.mask, [True, False, False, False, False, False, False, False, True, True])


def test_set_time():
    for x in [get_ndvar(2, 100, 0), get_ndvar(2, 100, 8)]:
        x_sub = x.sub(time=(0.000, None))
        assert x_sub.time.tmin == 0.000
        x_pad = set_time(x_sub, x)
        assert x_pad.time.tmin == -0.100
        assert x_pad.x.ravel()[0] == 0
        x_pad = set_time(x_sub, x, mode='edge')
        assert x_pad.time.tmin == -0.100
        assert x_pad.x.ravel()[0] == x_sub.x.ravel()[0]


def test_smoothing():
    x = get_ndvar(2)
    xt = NDVar(x.x.swapaxes(1, 2), [x.dims[i] for i in [0, 2, 1]], x.name, x.info)

    # smoothing across time
    ma = x.smooth('time', 0.2, 'blackman')
    assert_dataobj_equal(x.smooth('time', window='blackman', window_samples=20), ma)
    with pytest.raises(TypeError):
        x.smooth('time')
    with pytest.raises(TypeError):
        x.smooth('time', 0.2, 'blackman', window_samples=20)
    mas = xt.smooth('time', 0.2, 'blackman')
    assert_allclose(ma.x, mas.x.swapaxes(1, 2), 1e-10)
    ma_mean = x.mean('case').smooth('time', 0.2, 'blackman')
    assert_allclose(ma.mean('case').x, ma_mean.x)
    # against raw scipy.signal
    window = signal.get_window('blackman', 20, False)
    window /= window.sum()
    window.shape = (1, 20, 1)
    assert_array_equal(ma.x[:, 10:-10], signal.convolve(x.x, window, 'same')[:, 10:-10])
    # mode parameter
    full = signal.convolve(x.x, window, 'full')
    ma = x.smooth('time', 0.2, 'blackman', mode='left')
    assert_array_equal(ma.x[:], full[:, :100])
    ma = x.smooth('time', 0.2, 'blackman', mode='right')
    assert_array_equal(ma.x[:], full[:, 19:])

    # fix_edges: smooth with constant sum
    xs = x.smooth('frequency', window_samples=1, fix_edges=True)
    assert_dataobj_equal(xs.sum('frequency'), x.sum('frequency'))
    xs = x.smooth('frequency', window_samples=2, fix_edges=True)
    assert_dataobj_equal(xs.sum('frequency'), x.sum('frequency'), 14)
    xs = x.smooth('frequency', window_samples=3, fix_edges=True)
    assert_dataobj_equal(xs.sum('frequency'), x.sum('frequency'), 14)
    xs = x.smooth('frequency', window_samples=5, fix_edges=True)
    assert_dataobj_equal(xs.sum('frequency'), x.sum('frequency'), 14)
    xs = x.smooth('frequency', window_samples=4, fix_edges=True)
    assert_dataobj_equal(xs.sum('frequency'), x.sum('frequency'), 14)

    # gaussian
    x = get_ndvar(2, frequency=0, sensor=5)
    x.smooth('sensor', 0.1, 'gaussian')
    x = get_ndvar(2, sensor=5)
    x.smooth('sensor', 0.1, 'gaussian')



"""Plot non-uniform time series"""


class PltBinNuts:
    def __init__(self, ax, epoch, color='r', fill=False, hatch='//', **kwargs):
        """Plot a simple on/off nonuniform time series

        Parameters
        ----------
        ax : matplotlib axes
            Target axes.
        epoch : array
            Array with fields 'start' and 'stop'.
        kwargs :
            axvspan keyword arguments.
        """
        self._handles = []
        for line in epoch:
            start = line['start']
            stop = line['stop']
            h = ax.axvspan(start, stop, color=color, fill=fill, hatch=hatch, **kwargs)
            self._handles.append(h)



# skip test: data unavailable
from pathlib import Path


def iter_top_subject_dirs(root: Path):
    for p in sorted(root.iterdir()):
        if p.is_dir() and p.name.startswith("sub-"):
            yield p


def plan_bem_file_renames(subj_dir: Path):
    bem = subj_dir / "bem"
    if not bem.is_dir():
        return
    for f in sorted(bem.iterdir()):
        if f.is_file() and f.name.startswith("sub-"):
            yield f, f.with_name(f.name[4:])


def plan_dir_rename(subj_dir: Path):
    name = subj_dir.name
    if name.startswith("sub-"):
        target = subj_dir.with_name(name[4:])
        return subj_dir, target
    return None


def main(root: str | Path, dry_run: bool = False):
    """Run the rename routine (see module docstring)."""
    root_path = Path(root).expanduser().resolve()

    if not root_path.is_dir():
        print(f"[ERROR] Root directory is not accessible: {root_path}")
        return

    bem_plans = []
    dir_plans = []

    # Stage 1: BEM files
    for subj in iter_top_subject_dirs(root_path):
        for old, new in plan_bem_file_renames(subj):
            if not new.exists():
                bem_plans.append((old, new))
            else:
                print(f"[SKIP] Conflict (file exists): {new}")

    # Stage 2: Directories
    for subj in iter_top_subject_dirs(root_path):
        plan = plan_dir_rename(subj)
        if plan:
            old, new = plan
            if not new.exists():
                dir_plans.append((old, new))
            else:
                print(f"[SKIP] Conflict (directory exists): {new}")

    if not bem_plans and not dir_plans:
        print("No items to rename.")
        return 0

    # Plan output
    if bem_plans:
        print("=== BEM file rename plan ===")
        for o, n in bem_plans:
            print(f"FILE: {o.name} -> {n.name}")
    if dir_plans:
        print("=== Directory rename plan (after BEM files) ===")
        for o, n in dir_plans:
            print(f"DIR:  {o.name} -> {n.name}")
    if dry_run:
        print("\nDry-run mode: no changes made.")
        return

    # Execute: files first (BEM), then directories
    for o, n in bem_plans:
        try:
            o.rename(n)
            print(f"[OK] FILE {o.name} -> {n.name}")
        except Exception as e:
            print(f"[ERR] {o} -> {n}: {e}")

    for o, n in dir_plans:
        try:
            o.rename(n)
            print(f"[OK] DIR  {o.name} -> {n.name}")
        except Exception as e:
            print(f"[ERR] {o} -> {n}: {e}")


if __name__ == "__main__":
    # Set dry_run=True to only show the planned rename operations without making changes.
    main('/mnt/d/Data/ds005810/derivatives/freesurfer', dry_run=False)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain import *

# turn off multiprocessing
configure(0)
ds = datasets._get_continuous(10000)

timeit boosting(ds['y'], ds['x1'], 0, .5)



import wx

from .._utils import IS_OSX
from . import ID

FOCUS_UI_UPDATE_FUNC_NAMES = {
    wx.ID_COPY: 'CanCopy',
    ID.COPY_AS_PNG: 'CanCopyPNG',
    wx.ID_CUT: 'CanCut',
    wx.ID_PASTE: 'CanPaste',
}


class EelbrainWindow:
    # Frame subclass to support UI Update

    def OnWindowIconize(self, event):
        self.Iconize()

    def OnWindowZoom(self, event):
        self.Maximize()

    def OnUpdateUIBackward(self, event):
        event.Enable(False)

    def OnUpdateUIClear(self, event):
        event.Enable(False)

    def OnUpdateUIClose(self, event):
        event.Enable(True)

    def OnUpdateUIDown(self, event):
        event.Enable(True)

    def OnUpdateUIDrawCrosshairs(self, event):
        event.Enable(False)
        event.Check(False)

    def OnUpdateUIForward(self, event):
        event.Enable(False)

    def OnUpdateUIFocus(self, event):
        func_name = FOCUS_UI_UPDATE_FUNC_NAMES[event.GetId()]
        win = self.FindFocus()
        func = getattr(win, func_name, None)
        if func is None:
            func = getattr(self, func_name, None)
            if func is None:
                event.Enable(False)
                return
        event.Enable(func())

    def OnUpdateUIOpen(self, event):
        event.Enable(False)

    def OnUpdateUIRedo(self, event):
        event.Enable(False)

    def OnUpdateUISave(self, event):
        event.Enable(False)

    def OnUpdateUISaveAs(self, event):
        event.Enable(False)

    def OnUpdateUISetLayout(self, event):
        event.Enable(False)

    def OnUpdateUISetMarkedChannels(self, event):
        event.Enable(False)

    def OnUpdateUISetVLim(self, event):
        event.Enable(False)

    def OnUpdateUISetTime(self, event):
        event.Enable(False)

    def OnUpdateUITools(self, event):
        event.Enable(hasattr(self, 'MakeToolsMenu'))

    def OnUpdateUIUndo(self, event):
        event.Enable(False)

    def OnUpdateUIUp(self, event):
        event.Enable(False)


class EelbrainFrame(wx.Frame, EelbrainWindow):
    _allow_user_set_title = False

    def __init__(self, parent=None, id=wx.ID_ANY, title="", pos=wx.DefaultPosition, *args, **kwargs):
        wx.Frame.__init__(self, parent, id, title, pos, *args, **kwargs)
        if not IS_OSX:
            from .app import get_app
            self.SetMenuBar(get_app().CreateMenu(self))
        self._title = self.GetTitle()

    def OnClear(self, event):
        raise RuntimeError(str(self))

    def OnCopy(self, event):
        win = wx.Window.FindFocus()
        if hasattr(win, 'CanCopy'):
            return win.Copy()
        elif hasattr(self, 'CanCopy'):
            return self.Copy()
        else:
            event.Skip()

    def OnDrawCrosshairs(self, event):
        raise RuntimeError(str(self))

    def OnOpen(self, event):
        raise RuntimeError(str(self))

    def OnRedo(self, event):
        raise RuntimeError(str(self))

    def OnSave(self, event):
        raise RuntimeError(str(self))

    def OnSaveAs(self, event):
        raise RuntimeError(str(self))

    def OnSetVLim(self, event):
        raise RuntimeError(str(self))

    def OnSetLayout(self, event):
        raise RuntimeError(str(self))

    def OnSetMarkedChannels(self, event):
        raise RuntimeError(str(self))

    def OnSetTime(self, event):
        raise RuntimeError(str(self))

    def OnSetWindowTitle(self, event):
        dlg = wx.TextEntryDialog(self, f"New title for '{self._title}':", "Set Window Title", value=self._title)
        if dlg.ShowModal() == wx.ID_OK:
            self._title = dlg.GetValue()
            self.SetTitle(self._title)
        dlg.Destroy()

    def OnUndo(self, event):
        raise RuntimeError(str(self))

    def OnWindowClose(self, event):
        self.Close()

    def SetTitleSuffix(self, suffix):
        self.SetTitle(self._title + suffix)


class EelbrainDialog(wx.Dialog, EelbrainWindow):

    pass
